#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.ripemd160 import ripemd160
from ..libs.bech32 import (
    bech32_encode, bech32_decode
)
from ..eccs import (
    IPublicKey, SLIP10Secp256k1PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Cosmos
from ..crypto import sha256
from ..utils import bytes_to_string
from .iaddress import IAddress


class CosmosAddress(IAddress):

    hrp: str = Cosmos.NETWORKS.MAINNET.HRP

    @staticmethod
    def name() -> str:
        """
        Returns the name of the cryptocurrency or network.

        :return: The name "Cosmos".
        :rtype: str
        """

        return "Cosmos"

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encodes the given public key into a Bech32 Cosmos address.

        :param public_key: The public key to be encoded.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments, including "hrp".
        :type kwargs: Any

        :return: The encoded Bech32 Cosmos address.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Secp256k1PublicKey
        )
        public_key_hash: bytes = ripemd160(sha256(
            public_key.raw_compressed()
        ))
        return bech32_encode(
            kwargs.get("hrp", cls.hrp), public_key_hash
        )

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decodes the given Bech32 Cosmos address into its corresponding public key hash.

        :param address: The Bech32 Cosmos address to be decoded.
        :type address: str
        :param kwargs: Additional keyword arguments, including "hrp".
        :type kwargs: Any

        :return: The decoded public key hash.
        :rtype: str
        """

        hrp, address_decode = bech32_decode(
            kwargs.get("hrp", cls.hrp), address
        )
        return bytes_to_string(address_decode)



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Type
from bip38 import BIP38

import json
import click
import sys

from ..entropies import ENTROPIES
from ..mnemonics import MNEMONICS
from ..seeds import SEEDS
from ..hds import HDS
from ..derivations import DERIVATIONS
from ..cryptocurrencies import (
    ICryptocurrency, get_cryptocurrency
)

from ..hdwallet import HDWallet
from . import BIP38_CRYPTOCURRENCIES


def dump(**kwargs) -> None:
    try:
        cryptocurrency: Type[ICryptocurrency] = get_cryptocurrency(
            symbol=kwargs.get("symbol")
        )
        if not HDS.is_hd(name=kwargs.get("hd")):
            click.echo(click.style(
                f"Wrong HD name, (expected={HDS.names()}, got='{kwargs.get('hd')}')"
            ), err=True)
            sys.exit()
        if not DERIVATIONS.is_derivation(name=kwargs.get("derivation")):
            click.echo(click.style(
                f"Wrong from derivation name, (expected={DERIVATIONS.names()}, got='{kwargs.get('derivation')}')"
            ), err=True)
            sys.exit()
        if not cryptocurrency.NETWORKS.is_network(network=kwargs.get("network")):
            click.echo(click.style(
                f"Invalid {cryptocurrency.NAME} cryptocurrency network, "
                f"(expected={cryptocurrency.NETWORKS.get_networks()}, got='{kwargs.get('network')}')"
            ), err=True)
            sys.exit()

        semantic = kwargs.get("semantic")
        if semantic is None:
            if kwargs.get("hd") in [
                "BIP32", "BIP44", "BIP86", "Cardano"
            ]:
                semantic = cryptocurrency.DEFAULT_SEMANTIC
            elif kwargs.get("hd") == "BIP49":
                semantic = "p2wpkh-in-p2sh"
            elif kwargs.get("hd") in ["BIP84", "BIP141"]:
                semantic = "p2wpkh"
                
        hdwallet: HDWallet = HDWallet(
            cryptocurrency=cryptocurrency,
            hd=HDS.hd(name=kwargs.get("hd")),
            network=kwargs.get("network"),
            public_key_type=kwargs.get("public_key_type"),
            language=kwargs.get("language"),
            passphrase=kwargs.get("passphrase"),
            cardano_type=kwargs.get("cardano_type"),
            address_type=kwargs.get("address_type"),
            staking_public_key=kwargs.get("staking_public_key"),
            mode=kwargs.get("mode"),
            mnemonic_type=kwargs.get("mnemonic_type"),
            checksum=kwargs.get("checksum"),
            payment_id=kwargs.get("payment_id"),
            semantic=semantic
        )

        if kwargs.get("entropy"):
            if not ENTROPIES.is_entropy(name=kwargs.get("entropy_client")):
                click.echo(click.style(
                    f"Wrong entropy client, (expected={ENTROPIES.names()}, got='{kwargs.get('entropy_client')}')"
                ), err=True)
                sys.exit()
            hdwallet.from_entropy(
                entropy=ENTROPIES.entropy(name=kwargs.get("entropy_client")).__call__(
                    entropy=kwargs.get("entropy")
                )
            )
        elif kwargs.get("mnemonic"):
            if not MNEMONICS.is_mnemonic(name=kwargs.get("mnemonic_client")):
                click.echo(click.style(
                    f"Wrong mnemonic client, (expected={MNEMONICS.names()}, got='{kwargs.get('mnemonic_client')}')"
                ), err=True)
                sys.exit()
            if kwargs.get("mnemonic_client") == "Electrum-V2":
                hdwallet.from_mnemonic(
                    mnemonic=MNEMONICS.mnemonic(name=kwargs.get("mnemonic_client")).__call__(
                        mnemonic=kwargs.get("mnemonic"),
                        mnemonic_type=kwargs.get("mnemonic_type")
                    )
                )
            else:
                hdwallet.from_mnemonic(
                    mnemonic=MNEMONICS.mnemonic(name=kwargs.get("mnemonic_client")).__call__(
                        mnemonic=kwargs.get("mnemonic")
                    )
                )
        elif kwargs.get("seed"):
            if not SEEDS.is_seed(name=kwargs.get("seed_client")):
                click.echo(click.style(
                    f"Wrong seed client, (expected={SEEDS.names()}, got='{kwargs.get('seed_client')}')"
                ), err=True)
                sys.exit()
            hdwallet.from_seed(
                seed=SEEDS.seed(name=kwargs.get("seed_client")).__call__(
                    seed=kwargs.get("seed")
                )
            )
        elif kwargs.get("xprivate_key"):
            hdwallet.from_xprivate_key(
                xprivate_key=kwargs.get("xprivate_key"),
                encoded=kwargs.get("encoded"),
                strict=kwargs.get("strict")
            )
        elif kwargs.get("xpublic_key"):
            hdwallet.from_xpublic_key(
                xpublic_key=kwargs.get("xpublic_key"),
                encoded=kwargs.get("encoded"),
                strict=kwargs.get("strict")
            )
        elif kwargs.get("private_key"):
            hdwallet.from_private_key(
                private_key=kwargs.get("private_key")
            )
        elif kwargs.get("wif"):

            _wif = kwargs.get("wif")

            if kwargs.get("bip38"):

                bip38: BIP38 = BIP38(
                  cryptocurrency=BIP38_CRYPTOCURRENCIES[cryptocurrency.NAME], network=kwargs.get("network")
                )
                _wif = bip38.decrypt(encrypted_wif=_wif, passphrase=kwargs.get("passphrase"))

            hdwallet.from_wif(
                wif=_wif
            )

        elif kwargs.get("public_key"):
            hdwallet.from_public_key(
                public_key=kwargs.get("public_key")
            )
        elif kwargs.get("spend_private_key"):
            hdwallet.from_spend_private_key(
                spend_private_key=kwargs.get("spend_private_key")
            )
        elif kwargs.get("view_private_key") and kwargs.get("spend_public_key"):
            hdwallet.from_watch_only(
                view_private_key=kwargs.get("view_private_key"),
                spend_public_key=kwargs.get("spend_public_key")
            )

        if (
            kwargs.get("entropy") or
            kwargs.get("mnemonic") or
            kwargs.get("seed") or
            kwargs.get("xprivate_key") or
            kwargs.get("xpublic_key") or
            (kwargs.get("private_key") and kwargs.get("hd") in ["Electrum-V1", "Monero"]) or
            (kwargs.get("wif") and kwargs.get("hd") == "Electrum-V1") or
            kwargs.get("spend_private_key") or
            (kwargs.get("view_private_key") and kwargs.get("spend_public_key"))
        ):

            if kwargs.get("derivation") in [
                "BIP44", "BIP49", "BIP84", "BIP86"
            ]:
                hdwallet.from_derivation(
                    derivation=DERIVATIONS.derivation(name=kwargs.get("derivation")).__call__(
                        coin_type=cryptocurrency.COIN_TYPE,
                        account=kwargs.get("account"),
                        change=kwargs.get("change"),
                        address=kwargs.get("address")
                    )
                )
            elif kwargs.get("derivation") == "CIP1852":
                hdwallet.from_derivation(
                    derivation=DERIVATIONS.derivation(name=kwargs.get("derivation")).__call__(
                        coin_type=cryptocurrency.COIN_TYPE,
                        account=kwargs.get("account"),
                        role=kwargs.get("role"),
                        address=kwargs.get("address")
                    )
                )
            elif kwargs.get("derivation") == "Custom":
                hdwallet.from_derivation(
                    derivation=DERIVATIONS.derivation(name=kwargs.get("derivation")).__call__(
                        path=kwargs.get("path", "m/"),
                        indexes=kwargs.get("indexes", [])
                    )
                )
            elif kwargs.get("derivation") == "Electrum":
                hdwallet.from_derivation(
                    derivation=DERIVATIONS.derivation(name=kwargs.get("derivation")).__call__(
                        change=kwargs.get("change"),
                        address=kwargs.get("address")
                    )
                )
            elif kwargs.get("derivation") == "Monero":
                hdwallet.from_derivation(
                    derivation=DERIVATIONS.derivation(name=kwargs.get("derivation")).__call__(
                        minor=kwargs.get("minor"),
                        major=kwargs.get("major")
                    )
                )

        click.echo(json.dumps(
            hdwallet.dump(exclude=set(kwargs.get("exclude").split(","))), indent=4, ensure_ascii=False
        ))

    except Exception as exception:
        click.echo(click.style(
            f"Error: {str(exception)}"
        ), err=True)
        sys.exit()



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

import pytest
import re

from hdwallet.keys import (
    is_valid_key, is_root_key
)
from hdwallet.exceptions import ExtendedKeyError


def test_keys():

    assert is_valid_key("xprv9s21ZrQH143K3hU8mSCiGTDNX8vESS6bVuXKpMgJK8aLfWTsCgakwRfCts6bXoRG51sHYenkyseTeWB12RfS8KgQF7w8e8ner4U7HDBDuMw")
    assert is_root_key("xprv9s21ZrQH143K3hU8mSCiGTDNX8vESS6bVuXKpMgJK8aLfWTsCgakwRfCts6bXoRG51sHYenkyseTeWB12RfS8KgQF7w8e8ner4U7HDBDuMw")
    assert not is_root_key("xprvA31rahaYRvc6tpxRoBUiymeY5k6c2uCQwWtQxRXsU4cX7FmNyL3fg3T9mXyYc2k7huWwaM3Vi19n6tQ21VgXH65Ws7Snz681y3AmqXrNCL1")
    assert is_valid_key("xpub661MyMwAqRbcGBYbsTjidbA75AkiqtpSs8Svck5usU7KYJo1kDu1VDygk8tWBWC2hvmdFBFEwuz1W2aPHsqKjRcye6MeeyDFkRPNbUq73N9")
    assert is_root_key("xpub661MyMwAqRbcGBYbsTjidbA75AkiqtpSs8Svck5usU7KYJo1kDu1VDygk8tWBWC2hvmdFBFEwuz1W2aPHsqKjRcye6MeeyDFkRPNbUq73N9")
    assert not is_root_key("xpub6GAzPQuoyqr6sT251c2dqZkcYU9FZZpvtiJsnADjBxU6yaF577vE6pC9Lyb3ReSLWZsMQMtcVoPKV1qorppQWbTsxwA7r1AgZ9PbfT1iULz")

    assert is_valid_key("xprv3QESAWYc9vDdZiM9o6Z4Dsny5f4zBVdWgdU2YKw25fGyhT8ZgEkkta4kmkFhn875Mq3xUjRaGyhCnFoK1qb7fDELTyCJMTdiatAHuvG8mjjJgfJesQ9AFBikt59YLVgTEn2HhXKYxjmKSvxHZxdQXzF")
    assert is_root_key("xprv3QESAWYc9vDdZiM9o6Z4Dsny5f4zBVdWgdU2YKw25fGyhT8ZgEkkta4kmkFhn875Mq3xUjRaGyhCnFoK1qb7fDELTyCJMTdiatAHuvG8mjjJgfJesQ9AFBikt59YLVgTEn2HhXKYxjmKSvxHZxdQXzF")
    assert not is_root_key("xprv3TFbDUeC4U4KfWsziG2VfvZ4C2b9foDhQcaFc7vgLCduXkWARCHRqPQZDLYvnrxVaYmXdzPyf7aMuA8BaavsBXaMCCs1c7QhoGe7gGHMMTB7uenMe4YBh5XDWFNwfL6754zdagnqMq4AgysHuV5jnEv")

    assert not is_valid_key("xprv9s21ZrQH143K3hU8mSCiGTDNX8vESS6bVuX.............enkyseTeWB12RfS8KgQF7w8e8ne4U7HDBDuMw")
    with pytest.raises(ExtendedKeyError, match=re.escape("Invalid extended(x) key")):
        is_root_key("xprv9s21ZrQH143K3hU8mSCiGTDNX8vESS6bVuX.............enkyseTeWB12RfS8KgQF7w8e8ne4U7HDBDuMw")



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.base32 import (
    encode_no_padding, decode
)
from ..eccs import (
    IPublicKey, SLIP10Ed25519Blake2bPublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Nano
from ..crypto import blake2b_40
from ..utils import (
    get_bytes, bytes_reverse, bytes_to_string
)
from .iaddress import IAddress


class NanoAddress(IAddress):

    address_prefix: str = Nano.PARAMS.ADDRESS_PREFIX
    alphabet: str = Nano.PARAMS.ALPHABET
    payload_padding_decoded: bytes = Nano.PARAMS.PAYLOAD_PADDING_DECODED
    payload_padding_encoded: str = Nano.PARAMS.PAYLOAD_PADDING_ENCODED

    @staticmethod
    def name() -> str:
        """
        Returns the name of the cryptocurrency, which is "Nano".

        :return: The name of the cryptocurrency.
        :rtype: str
        """

        return "Nano"

    @classmethod
    def compute_checksum(cls, public_key: bytes) -> bytes:
        return bytes_reverse(blake2b_40(public_key))

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encode a public key into an address using custom encoding.

        :param public_key: The public key to encode.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any

        :return: The encoded address.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key, SLIP10Ed25519Blake2bPublicKey
        )
        checksum: bytes = cls.compute_checksum(public_key.raw_compressed()[1:])
        payload: bytes = (
            cls.payload_padding_decoded + public_key.raw_compressed()[1:] + checksum
        )
        b32_encoded: str = encode_no_padding(bytes_to_string(payload), custom_alphabet=cls.alphabet)

        return cls.address_prefix + b32_encoded[len(cls.payload_padding_encoded):]

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decode an address string into a public key using custom decoding.

        :param address: The address string to decode.
        :type address: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any

        :return: The decoded public key as a string.
        :rtype: str
        """

        prefix_got: str = address[:len(cls.address_prefix)]
        if cls.address_prefix != prefix_got:
            raise ValueError(f"Invalid prefix (expected: {cls.address_prefix}, got: {prefix_got})")
        address_no_prefix: str = address[len(cls.address_prefix):]

        address_decode: bytes = get_bytes(decode(
            cls.payload_padding_encoded + address_no_prefix, custom_alphabet=cls.alphabet
        ))

        expected_length: int = (
            SLIP10Ed25519Blake2bPublicKey.compressed_length() + 5 + len(cls.payload_padding_decoded) - 1
        )
        if len(address_decode) != expected_length:
            raise ValueError(f"Invalid length (expected: {expected_length}, got: {len(address_decode)})")

        checksum: bytes = address_decode[len(cls.payload_padding_decoded):][-1 * 5:]
        public_key: bytes = address_decode[len(cls.payload_padding_decoded):][:-1 * 5]

        checksum_got: bytes = cls.compute_checksum(public_key)
        if checksum != checksum_got:
            raise ValueError(f"Invalid checksum (expected: {bytes_to_string(checksum)}, got: {bytes_to_string(checksum_got)})")

        if not SLIP10Ed25519Blake2bPublicKey.is_valid_bytes(public_key):
            raise ValueError(f"Invalid {SLIP10Ed25519Blake2bPublicKey.name()} public key {bytes_to_string(public_key)}")

        return bytes_to_string(public_key)



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Type, List, Tuple
)
from bip38 import BIP38

import json
import click
import sys
import csv

from ..entropies import ENTROPIES
from ..mnemonics import MNEMONICS
from ..seeds import SEEDS
from ..hds import (
    BIP32HD, BIP44HD, BIP49HD, BIP84HD, BIP86HD, BIP141HD, CardanoHD, ElectrumV1HD, ElectrumV2HD, MoneroHD, HDS
)
from ..derivations import (
    IDerivation, DERIVATIONS
)
from ..cryptocurrencies import (
    ICryptocurrency, get_cryptocurrency
)
from ..hdwallet import HDWallet
from . import BIP38_CRYPTOCURRENCIES


def dumps(**kwargs) -> None:
    try:
        cryptocurrency: Type[ICryptocurrency] = get_cryptocurrency(
            symbol=kwargs.get("symbol")
        )
        if not HDS.is_hd(name=kwargs.get("hd")):
            click.echo(click.style(
                f"Wrong HD name, (expected={HDS.names()}, got='{kwargs.get('hd')}')"
            ), err=True)
            sys.exit()
        if not DERIVATIONS.is_derivation(name=kwargs.get("derivation")):
            click.echo(click.style(
                f"Wrong from derivation name, (expected={DERIVATIONS.names()}, got='{kwargs.get('derivation')}')"
            ), err=True)
            sys.exit()
        if not cryptocurrency.NETWORKS.is_network(network=kwargs.get("network")):
            click.echo(click.style(
                f"Invalid {cryptocurrency.NAME} cryptocurrency network, "
                f"(expected={cryptocurrency.NETWORKS.get_networks()}, got='{kwargs.get('network')}')"
            ), err=True)
            sys.exit()

        semantic = kwargs.get("semantic")
        if semantic is None:
            if kwargs.get("hd") in [
                "BIP32", "BIP44", "BIP86", "Cardano"
            ]:
                semantic = cryptocurrency.DEFAULT_SEMANTIC
            elif kwargs.get("hd") == "BIP49":
                semantic = "p2wpkh-in-p2sh"
            elif kwargs.get("hd") in ["BIP84", "BIP141"]:
                semantic = "p2wpkh"

        hdwallet: HDWallet = HDWallet(
            cryptocurrency=cryptocurrency,
            hd=HDS.hd(name=kwargs.get("hd")),
            network=kwargs.get("network"),
            public_key_type=kwargs.get("public_key_type"),
            language=kwargs.get("language"),
            passphrase=kwargs.get("passphrase"),
            cardano_type=kwargs.get("cardano_type"),
            address_type=kwargs.get("address_type"),
            staking_public_key=kwargs.get("staking_public_key"),
            mode=kwargs.get("mode"),
            mnemonic_type=kwargs.get("mnemonic_type"),
            checksum=kwargs.get("checksum"),
            payment_id=kwargs.get("payment_id"),
            semantic=semantic
        )

        if kwargs.get("entropy"):
            if not ENTROPIES.is_entropy(name=kwargs.get("entropy_client")):
                click.echo(click.style(
                    f"Wrong entropy client, (expected={ENTROPIES.names()}, got='{kwargs.get('entropy_client')}')"
                ), err=True)
                sys.exit()
            hdwallet.from_entropy(
                entropy=ENTROPIES.entropy(name=kwargs.get("entropy_client")).__call__(
                    entropy=kwargs.get("entropy")
                )
            )
        elif kwargs.get("mnemonic"):
            if not MNEMONICS.is_mnemonic(name=kwargs.get("mnemonic_client")):
                click.echo(click.style(
                    f"Wrong mnemonic client, (expected={MNEMONICS.names()}, got='{kwargs.get('mnemonic_client')}')"
                ), err=True)
                sys.exit()
            if kwargs.get("mnemonic_client") == "Electrum-V2":
                hdwallet.from_mnemonic(
                    mnemonic=MNEMONICS.mnemonic(name=kwargs.get("mnemonic_client")).__call__(
                        mnemonic=kwargs.get("mnemonic"),
                        mnemonic_type=kwargs.get("mnemonic_type")
                    )
                )
            else:
                hdwallet.from_mnemonic(
                    mnemonic=MNEMONICS.mnemonic(name=kwargs.get("mnemonic_client")).__call__(
                        mnemonic=kwargs.get("mnemonic")
                    )
                )
        elif kwargs.get("seed"):
            if not SEEDS.is_seed(name=kwargs.get("seed_client")):
                click.echo(click.style(
                    f"Wrong seed client, (expected={SEEDS.names()}, got='{kwargs.get('seed_client')}')"
                ), err=True)
                sys.exit()
            hdwallet.from_seed(
                seed=SEEDS.seed(name=kwargs.get("seed_client")).__call__(
                    seed=kwargs.get("seed")
                )
            )
        elif kwargs.get("xprivate_key"):
            hdwallet.from_xprivate_key(
                xprivate_key=kwargs.get("xprivate_key"),
                encoded=kwargs.get("encoded"),
                strict=kwargs.get("strict")
            )
        elif kwargs.get("xpublic_key"):
            hdwallet.from_xpublic_key(
                xpublic_key=kwargs.get("xpublic_key"),
                encoded=kwargs.get("encoded"),
                strict=kwargs.get("strict")
            )
        elif kwargs.get("private_key"):
            hdwallet.from_private_key(
                private_key=kwargs.get("private_key")
            )
        elif kwargs.get("wif"):
            _wif = kwargs.get("wif")

            if kwargs.get("bip38"):

                bip38: BIP38 = BIP38(
                  cryptocurrency=BIP38_CRYPTOCURRENCIES[cryptocurrency.NAME], network=kwargs.get("network")
                )
                _wif = bip38.decrypt(encrypted_wif=_wif, passphrase=kwargs.get("passphrase"))

            hdwallet.from_wif(
                wif=_wif
            )
        elif kwargs.get("public_key"):
            hdwallet.from_public_key(
                public_key=kwargs.get("public_key")
            )
        elif kwargs.get("spend_private_key"):
            hdwallet.from_spend_private_key(
                spend_private_key=kwargs.get("spend_private_key")
            )
        elif kwargs.get("view_private_key") and kwargs.get("spend_public_key"):
            hdwallet.from_watch_only(
                view_private_key=kwargs.get("view_private_key"),
                spend_public_key=kwargs.get("spend_public_key")
            )

        if (
            kwargs.get("entropy") or
            kwargs.get("mnemonic") or
            kwargs.get("seed") or
            kwargs.get("xprivate_key") or
            kwargs.get("xpublic_key") or
            (kwargs.get("private_key") and kwargs.get("hd") in ["Electrum-V1", "Monero"]) or
            (kwargs.get("wif") and kwargs.get("hd") == "Electrum-V1") or
            kwargs.get("spend_private_key") or
            (kwargs.get("view_private_key") and kwargs.get("spend_public_key"))
        ):

            if kwargs.get("derivation") in [
                "BIP44", "BIP49", "BIP84", "BIP86"
            ]:
                hdwallet.from_derivation(
                    derivation=DERIVATIONS.derivation(name=kwargs.get("derivation")).__call__(
                        coin_type=cryptocurrency.COIN_TYPE,
                        account=kwargs.get("account"),
                        change=kwargs.get("change"),
                        address=kwargs.get("address")
                    )
                )
            elif kwargs.get("derivation") == "CIP1852":
                hdwallet.from_derivation(
                    derivation=DERIVATIONS.derivation(name=kwargs.get("derivation")).__call__(
                        coin_type=cryptocurrency.COIN_TYPE,
                        account=kwargs.get("account"),
                        role=kwargs.get("role"),
                        address=kwargs.get("address")
                    )
                )
            elif kwargs.get("derivation") == "Custom":
                hdwallet.from_derivation(
                    derivation=DERIVATIONS.derivation(name=kwargs.get("derivation")).__call__(
                        path=kwargs.get("path", "m/"),
                        indexes=kwargs.get("indexes", [])
                    )
                )
            elif kwargs.get("derivation") == "Electrum":
                hdwallet.from_derivation(
                    derivation=DERIVATIONS.derivation(name=kwargs.get("derivation")).__call__(
                        change=kwargs.get("change"),
                        address=kwargs.get("address")
                    )
                )
            elif kwargs.get("derivation") == "Monero":
                hdwallet.from_derivation(
                    derivation=DERIVATIONS.derivation(name=kwargs.get("derivation")).__call__(
                        minor=kwargs.get("minor"),
                        major=kwargs.get("major")
                    )
                )
            elif kwargs.get("derivation") == "HDW":
                hdwallet.from_derivation(
                    derivation=DERIVATIONS.derivation(name=kwargs.get("derivation")).__call__(
                        account=kwargs.get("account"),
                        ecc=kwargs.get("ecc"),
                        address=kwargs.get("address")
                    )
                )


        hd_name: str = hdwallet._hd.name()
        if kwargs.get("include"):
            _include: str = kwargs.get("include")
        elif hd_name == BIP32HD.name():
            _include: str = "at:path,addresses:p2pkh,public_key,wif"
        elif hd_name in [
            BIP44HD.name(), BIP49HD.name(), BIP84HD.name(), BIP86HD.name(), BIP141HD.name()
        ]:
            _include: str = "at:path,address,public_key,wif"
        elif hd_name == CardanoHD.name():
            _include: str = "at:path,address,public_key,private_key"
        elif hd_name in [
            ElectrumV1HD.name(), ElectrumV2HD.name()
        ]:
            _include: str = "at:change,at:address,address,public_key,wif"
        elif hd_name == MoneroHD.name():
            _include: str = "at:minor,at:major,sub_address"
        else:
            raise Exception("Unknown HD")

        tmp_addresses = [
            "Algorand", "Aptos", "Avalanche", "Cosmos", "EOS", "Ergo", "Ethereum", "Filecoin", "Harmony", "Icon", "Injective", "MultiversX",
            "Nano", "Near", "Neo", "OKT-Chain", "Ripple", "Solana", "Stellar", "Sui", "Tezos", "Tron", "XinFin", "Zilliqa"
        ]
        tmp_cryptocurrency = get_cryptocurrency(symbol=hdwallet.symbol())
        if hdwallet.cryptocurrency() in ("Bitcoin-Cash", "Bitcoin-Cash-SLP", "eCash"):
            _include: str = "at:path,addresses:legacy-p2pkh,public_key,wif"
        elif any([address in tmp_addresses for address in tmp_cryptocurrency.ADDRESSES.get_addresses()]):
            _include: str = "at:path,address,public_key,private_key"

        if hdwallet.cryptocurrency() == "Avalanche":
            _include: str = "at:path,addresses:p-chain,public_key,wif"
        elif hdwallet.cryptocurrency() == "Binance":
            _include: str = "at:path,addresses:chain,public_key,wif"


        hdwallet_csv = csv.DictWriter(
            sys.stdout, fieldnames=_include.split(","), extrasaction="ignore", delimiter=kwargs.get("delimiter")
        )

        if kwargs.get("include_header"):
            hdwallet_csv.writeheader()

        def drive(*args) -> List[str]:
            def drive_helper(derivations, current_derivation: List[Tuple[int, bool]] = []) -> List[str]:
                if not derivations:

                    derivation_name: str = hdwallet._derivation.name()
                    if derivation_name in [
                        "BIP44", "BIP49", "BIP84", "BIP86"
                    ]:
                        _derivation: IDerivation = DERIVATIONS.derivation(
                            name=kwargs.get("derivation")
                        ).__call__(
                            coin_type=current_derivation[1][0],
                            account=current_derivation[2][0],
                            change=current_derivation[3][0],
                            address=current_derivation[4][0]
                        )
                    elif derivation_name == "CIP1852":
                        _derivation: IDerivation = DERIVATIONS.derivation(
                            name=kwargs.get("derivation")
                        ).__call__(
                            coin_type=current_derivation[1][0],
                            account=current_derivation[2][0],
                            role=current_derivation[3][0],
                            address=current_derivation[4][0]
                        )
                    elif derivation_name == 'Electrum':
                        _derivation: IDerivation = DERIVATIONS.derivation(
                            name=kwargs.get("derivation")
                        ).__call__(
                            change=current_derivation[0][0],
                            address=current_derivation[1][0]
                        )
                    elif derivation_name == "Monero":
                        _derivation: IDerivation = DERIVATIONS.derivation(
                            name=kwargs.get("derivation")
                        ).__call__(
                            minor=current_derivation[0][0],
                            major=current_derivation[1][0]
                        )
                    elif derivation_name == "HDW":
                        _derivation: IDerivation = DERIVATIONS.derivation(
                            name=kwargs.get("derivation")
                        ).__call__(
                            account=current_derivation[0][0],
                            ecc=current_derivation[1][0],
                            address=current_derivation[2][0]
                        )
                    else:
                        _derivation: IDerivation = DERIVATIONS.derivation(
                            name=kwargs.get("derivation")
                        ).__call__(
                            path="m/" + "/".join(
                                [str(item[0]) + "'" if item[1] else str(item[0]) for item in current_derivation]
                            )
                        )

                    hdwallet.update_derivation(
                        derivation=_derivation
                    )
                    if kwargs.get("format") == "csv":
                        new_dump: dict = { }
                        dump: dict = hdwallet.dump(exclude={"root"})
                        for key in [keys.split(":") for keys in _include.split(",")]:
                            if len(key) == 2:
                                new_dump.setdefault(f"{key[0]}:{key[1]}", dump[key[0]][key[1]])
                            else:
                                new_dump.setdefault(f"{key[0]}", dump[key[0]])
                        hdwallet_csv.writerow(new_dump)

                    elif kwargs.get("format") == "json":
                        excludes = kwargs.get("exclude").split(",")
                        dump: dict = hdwallet.dump(exclude={"root", *excludes})
                        click.echo(json.dumps(dump, indent=4, ensure_ascii=False))

                    return [_derivation.path()]

                path: List[str] = []
                if len(derivations[0]) == 3:
                    for value in range(derivations[0][0], derivations[0][1] + 1):
                        path += drive_helper(
                            derivations[1:], current_derivation + [(value, derivations[0][2])]
                        )
                else:
                    path += drive_helper(
                        derivations[1:], current_derivation + [derivations[0]]
                    )
                return path
            return drive_helper(args)

        if kwargs.get("format") == "csv":
            if hdwallet._derivation is None:
                return None

            drive(*hdwallet._derivation.derivations())

        elif kwargs.get("format") == "json":
            if hdwallet._derivation is None:
                return None

            excludes = kwargs.get("exclude").split(",")
            if "root" not in excludes:
                click.echo(json.dumps(
                    hdwallet.dump(exclude={'derivation', *excludes}), indent=4, ensure_ascii=False
                ))


            drive(*hdwallet._derivation.derivations())
        else:
            click.echo(click.style(
                f"Wrong format, (expected= json | csv, got='{kwargs.get('format')}')"
            ), err=True)
            sys.exit()

    except Exception as exception:
        click.echo(click.style(
            f"Error: {str(exception)}"
        ), err=True)
        sys.exit()



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Optional, List, Type
)

from ..eccs import IEllipticCurveCryptography
from ..consts import (
    Info, WitnessVersions, Entropies, Mnemonics, Seeds, HDs, Addresses, AddressTypes, AddressPrefixes, Networks, Params, XPrivateKeyVersions, XPublicKeyVersions
)


class INetwork:

    NAME: str
    # Bitcoin
    PUBLIC_KEY_ADDRESS_PREFIX: Optional[int] = None
    SCRIPT_ADDRESS_PREFIX: Optional[int] = None
    HRP: Optional[str] = None
    WITNESS_VERSIONS: Optional[WitnessVersions] = None
    XPRIVATE_KEY_VERSIONS: Optional[XPrivateKeyVersions] = None
    XPUBLIC_KEY_VERSIONS: Optional[XPublicKeyVersions] = None
    MESSAGE_PREFIX: Optional[str] = None
    WIF_PREFIX: Optional[int] = None
    # Bitcoin-Cash | Bitcoin-Cash-SLP | eCash
    LEGACY_PUBLIC_KEY_ADDRESS_PREFIX: Optional[int] = None
    STD_PUBLIC_KEY_ADDRESS_PREFIX: Optional[int] = None
    LEGACY_SCRIPT_ADDRESS_PREFIX: Optional[int] = None
    STD_SCRIPT_ADDRESS_PREFIX: Optional[int] = None
    # Monero
    STANDARD: Optional[int] = None
    INTEGRATED: Optional[int] = None
    SUB_ADDRESS: Optional[int] = None
    # Cardano
    TYPE: Optional[int] = None
    PAYMENT_ADDRESS_HRP: Optional[str] = None
    REWARD_ADDRESS_HRP: Optional[str] = None


class ICryptocurrency:

    NAME: str
    SYMBOL: str
    INFO: Info
    ECC: Type[IEllipticCurveCryptography]
    COIN_TYPE: int
    SUPPORT_BIP38: bool
    NETWORKS: Networks
    DEFAULT_NETWORK: INetwork
    ENTROPIES: Entropies
    MNEMONICS: Mnemonics
    SEEDS: Seeds
    HDS: HDs
    DEFAULT_HD: str
    DEFAULT_PATH: str
    ADDRESSES: Addresses
    DEFAULT_ADDRESS: str
    ADDRESS_TYPES: Optional[AddressTypes] = None
    DEFAULT_ADDRESS_TYPE: Optional[str] = None
    ADDRESS_PREFIXES: Optional[AddressPrefixes] = None
    DEFAULT_ADDRESS_PREFIX: Optional[str] = None
    SEMANTICS: List[str] = []
    DEFAULT_SEMANTIC: Optional[str] = None
    PARAMS: Optional[Params] = None



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.base58 import (
    ensure_string, encode, decode
)
from ..eccs import (
    IPublicKey, SLIP10Secp256k1PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Ergo
from ..crypto import blake2b_256
from ..utils import (
    integer_to_bytes, bytes_to_string
)
from ..exceptions import (
    AddressError, NetworkError
)
from .iaddress import IAddress


class ErgoAddress(IAddress):

    checksum_length: int = Ergo.PARAMS.CHECKSUM_LENGTH
    address_types: dict = {
        "p2pkh": Ergo.PARAMS.ADDRESS_TYPES.P2PKH,
        "p2sh": Ergo.PARAMS.ADDRESS_TYPES.P2SH
    }
    network_types: dict = {
        "mainnet": Ergo.NETWORKS.MAINNET.TYPE,
        "testnet": Ergo.NETWORKS.TESTNET.TYPE
    }

    @staticmethod
    def name() -> str:
        """
        Returns the name of the cryptocurrency.

        :return: The name of the cryptocurrency.
        :rtype: str
        """

        return "Ergo"

    @classmethod
    def compute_checksum(cls, public_key: bytes) -> bytes:
        return blake2b_256(public_key)[:cls.checksum_length]

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encodes the given public key into an Ergo address.

        :param public_key: The public key to encode.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments, including "address_type" and "network_type".
        :type kwargs: Any

        :return: The encoded Ergo address.
        :rtype: str
        """

        if not kwargs.get("address_type"):
            address_type: str = cls.address_types[Ergo.DEFAULT_ADDRESS_TYPE]
        else:
            if kwargs.get("address_type") not in Ergo.ADDRESS_TYPES.get_address_types():
                raise AddressError(
                    f"Invalid {cls.name()} address type",
                    expected=Ergo.ADDRESS_TYPES.get_address_types(),
                    got=kwargs.get("address_type")
                )
            address_type: str = cls.address_types[kwargs.get("address_type")]

        if not kwargs.get("network_type"):
            raise NetworkError(f"{cls.name()} network type is required")
        elif kwargs.get("network_type") not in Ergo.NETWORKS.get_networks():
            raise NetworkError(
                f"Invalid {cls.name()} network type",
                expected=Ergo.NETWORKS.get_networks(),
                got=kwargs.get("network_type")
            )
        network_type = cls.network_types[kwargs.get("network_type")]

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Secp256k1PublicKey
        )
        prefix: bytes = integer_to_bytes(address_type + network_type)
        address_payload: bytes = prefix + public_key.raw_compressed()
        checksum: bytes = cls.compute_checksum(address_payload)

        return ensure_string(encode(
            address_payload + checksum
        ))

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decodes the given Ergo address into its corresponding public key.

        :param address: The Ergo address to decode.
        :type address: str
        :param kwargs: Additional keyword arguments, including "address_type" and "network_type".
        :type kwargs: Any

        :return: The decoded public key.
        :rtype: str
        """

        if not kwargs.get("address_type"):
            address_type: str = cls.address_types[Ergo.DEFAULT_ADDRESS_TYPE]
        else:
            if kwargs.get("address_type") not in Ergo.ADDRESS_TYPES.get_address_types():
                raise AddressError(
                    f"Invalid {cls.name()} address type",
                    expected=Ergo.ADDRESS_TYPES.get_address_types(),
                    got=kwargs.get("address_type")
                )
            address_type: str = cls.address_types[kwargs.get("address_type")]

        if not kwargs.get("network_type"):
            raise NetworkError(f"{cls.name()} network type is required")
        elif kwargs.get("network_type") not in Ergo.NETWORKS.get_networks():
            raise NetworkError(
                f"Invalid {cls.name()} network type",
                expected=Ergo.NETWORKS.get_networks(),
                got=kwargs.get("network_type")
            )
        network_type = cls.network_types[kwargs.get("network_type")]

        prefix: bytes = integer_to_bytes(address_type + network_type)
        address_decode: bytes = decode(address)

        expected_length: int = SLIP10Secp256k1PublicKey.compressed_length() + cls.checksum_length + 1
        if len(address_decode) != expected_length:
            raise ValueError(f"Invalid length (expected: {expected_length}, got: {len(address_decode)})")

        checksum: bytes = address_decode[-1 * cls.checksum_length:]
        address_with_prefix: bytes = address_decode[:-1 * cls.checksum_length]

        checksum_got: bytes = cls.compute_checksum(address_with_prefix)
        if checksum != checksum_got:
            raise ValueError(f"Invalid checksum (expected: {checksum.hex()}, got: {checksum_got.hex()})")

        prefix_got: bytes = address_with_prefix[:len(prefix)]
        if prefix != prefix_got:
            raise ValueError(f"Invalid prefix (expected: {prefix}, got: {prefix_got})")
        public_key: bytes = address_with_prefix[len(prefix):]

        if not SLIP10Secp256k1PublicKey.is_valid_bytes(public_key):
            raise ValueError(f"Invalid {SLIP10Secp256k1PublicKey.name()} public key {public_key.hex()}")

        return bytes_to_string(public_key)



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from .....consts import SLIP10_ED25519_CONST
from ....iecc import IPoint
from .. import SLIP10Ed25519PublicKey
from .point import SLIP10Ed25519MoneroPoint


class SLIP10Ed25519MoneroPublicKey(SLIP10Ed25519PublicKey):

    @staticmethod
    def name() -> str:
        """
        Get the name of the ecc class.

        :return: The name of the ecc class.
        :rtype: str
        """

        return "SLIP10-Ed25519-Monero"

    @staticmethod
    def compressed_length() -> int:
        """
        Returns the compressed length of the Ed25519 Monero public key.

        :return: The compressed length of the Ed25519 Monero public key.
        :rtype: int
        """

        return SLIP10_ED25519_CONST.PUBLIC_KEY_BYTE_LENGTH

    @staticmethod
    def uncompressed_length() -> int:
        """
        Returns the uncompressed length of the Ed25519 Monero public key.

        :return: The uncompressed length of the Ed25519 Monero public key.
        :rtype: int
        """

        return SLIP10Ed25519MoneroPublicKey.compressed_length()

    def raw_compressed(self) -> bytes:
        """
        Retrieves the raw compressed public key bytes.

        :return: The raw compressed public key bytes.
        :rtype: bytes
        """

        return bytes(self.verify_key)

    def raw_uncompressed(self) -> bytes:
        """
        Retrieves the raw uncompressed public key bytes.

        :return: The raw compressed public key bytes.
        :rtype: bytes
        """

        return self.raw_compressed()

    def point(self) -> IPoint:
        """
        Retrieves the point on the elliptic curve corresponding to the public key.

        :return: The elliptic curve point corresponding to the public key.
        :rtype: IPoint
        """

        return SLIP10Ed25519MoneroPoint(bytes(self.verify_key))



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    List, Dict, Type
)

from .algorand import AlgorandHD
from .bip32 import BIP32HD
from .bip44 import BIP44HD
from .bip49 import BIP49HD
from .bip84 import BIP84HD
from .bip86 import BIP86HD
from .bip141 import BIP141HD
from .cardano import CardanoHD
from .electrum import (
    ElectrumV1HD, ElectrumV2HD
)
from .monero import MoneroHD
from ..exceptions import HDError
from .ihd import IHD


class HDS:
    """
    A class that manages a dictionary of hd classes.

    This class provides methods to retrieve names and classes of various hd implementations,
    as well as methods to validate and access specific hd classes by name.

    Here are available hd names and classes:

    +--------------+--------------------------------------------------+
    | Name         | Class                                            |
    +==============+==================================================+
    | Algorand     | :class:`hdwallet.hds.algorand.AlgorandHD`        |
    +--------------+--------------------------------------------------+
    | BIP32        | :class:`hdwallet.hds.bip32.BIP32HD`              |
    +--------------+--------------------------------------------------+
    | BIP44        | :class:`hdwallet.hds.bip44.BIP44HD`              |
    +--------------+--------------------------------------------------+
    | BIP49        | :class:`hdwallet.hds.bip49.BIP49HD`              |
    +--------------+--------------------------------------------------+
    | BIP84        | :class:`hdwallet.hds.bip84.BIP84HD`              |
    +--------------+--------------------------------------------------+
    | BIP86        | :class:`hdwallet.hds.bip86.BIP86HD`              |
    +--------------+--------------------------------------------------+
    | BIP141       | :class:`hdwallet.hds.bip141.BIP141HD`            |
    +--------------+--------------------------------------------------+
    | Cardano      | :class:`hdwallet.hds.cardano.CardanoHD`          |
    +--------------+--------------------------------------------------+
    | Electrum-V1  | :class:`hdwallet.hds.electrum.v1.ElectrumV1HD`   |
    +--------------+--------------------------------------------------+
    | Electrum-V2  | :class:`hdwallet.hds.electrum.v2.ElectrumV2HD`   |
    +--------------+--------------------------------------------------+
    | Monero       | :class:`hdwallet.hds.monero.MoneroHD`            |
    +--------------+--------------------------------------------------+

    """

    dictionary: Dict[str, Type[IHD]] = {
        AlgorandHD.name(): AlgorandHD,
        BIP32HD.name(): BIP32HD,
        BIP44HD.name(): BIP44HD,
        BIP49HD.name(): BIP49HD,
        BIP84HD.name(): BIP84HD,
        BIP86HD.name(): BIP86HD,
        BIP141HD.name(): BIP141HD,
        CardanoHD.name(): CardanoHD,
        ElectrumV1HD.name(): ElectrumV1HD,
        ElectrumV2HD.name(): ElectrumV2HD,
        MoneroHD.name(): MoneroHD
    }

    @classmethod
    def names(cls) -> List[str]:
        """
        Get a list of names from the dictionary attribute of the class.

        :return: List of names from the class dictionary.
        :rtype: List[str]
        """

        return list(cls.dictionary.keys())

    @classmethod
    def classes(cls) -> List[Type[IHD]]:
        """
        Get a list of classes from the dictionary attribute of the class.

        :return: List of classes from the class dictionary.
        :rtype: List[Type[IHD]]
        """

        return list(cls.dictionary.values())

    @classmethod
    def hd(cls, name: str) -> Type[IHD]:
        """
        Get the HD class type from the dictionary attribute of the class based on the given name.

        :param name: The name of the HD class to retrieve.
        :type name: str

        :return: The HD class type corresponding to the given name.
        :rtype: Type[IHD]
        """

        if not cls.is_hd(name=name):
            raise HDError(
                "Invalid HD name", expected=cls.names(), got=name
            )

        return cls.dictionary[name]

    @classmethod
    def is_hd(cls, name: str) -> bool:
        """
        Check if the given name corresponds to an HD class in the class dictionary.

        :param name: The name to check.
        :type name: str

        :return: True if the name corresponds to an HD class, False otherwise.
        :rtype: bool
        """
        return name in cls.names()


__all__: List[str] = [
    "IHD", "HDS"
] + [
    cls.__name__ for cls in HDS.classes()
]



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Optional, Any
)

from ....libs.ed25519 import (
    point_add,
    int_encode,
    point_encode,
    point_is_encoded_bytes,
    point_is_generator,
    point_is_on_curve,
    point_is_decoded_bytes,
    point_coord_to_bytes,
    point_bytes_to_coord,
    point_scalar_mul_base,
    point_scalar_mul
)
from ...iecc import IPoint


class SLIP10Ed25519Point(IPoint):

    is_generator: bool
    point: bytes
    _x: Optional[int]
    _y: Optional[int]

    def __init__(self, point: bytes) -> None:
        """
        Initializes an instance with a byte representation of a point.

        :param point: The byte representation of the point.
        :type point: bytes
        """

        if not point_is_encoded_bytes(point):
            raise ValueError("Invalid point bytes")

        self.point = point
        self.is_generator = point_is_generator(point)
        self._x, self._y = None, None

    @staticmethod
    def name() -> str:
        """
        Get the name of the ecc class.

        :return: The name of the ecc class.
        :rtype: str
        """

        return "SLIP10-Ed25519"

    @classmethod
    def from_bytes(cls, point: bytes) -> IPoint:
        """
        Creates an instance of the Ed25519 point from its byte representation.

        :param point: The byte representation of the point.
        :type point: bytes

        :return: An instance of the Ed25519 point.
        :rtype: IPoint
        """

        if not point_is_on_curve(point):
            raise ValueError("Invalid point bytes")
        if point_is_decoded_bytes(point):
            point = point_encode(
                point_bytes_to_coord(point)
            )
        return cls(point)

    @classmethod
    def from_coordinates(cls, x: int, y: int) -> IPoint:
        """
        Creates an instance of the Ed25519 point from x and y coordinates.

        :param x: The x-coordinate of the point.
        :type x: int

        :param y: The y-coordinate of the point.
        :type y: int

        :return: An instance of the Ed25519 point.
        :rtype: IPoint
        """

        return cls.from_bytes(
            point_coord_to_bytes((x, y))
        )

    def underlying_object(self) -> Any:
        """
        Returns the underlying object representing the Ed25519 Monero point.

        :return: The underlying object representing the Ed25519 Monero point.
        :rtype: Any
        """

        return self.point

    def x(self) -> int:
        """
        Returns the x-coordinate of the Ed25519 Monero public key point.

        :return: The x-coordinate of the Ed25519 Monero public key point.
        :rtype: int
        """

        if self._x is None:
            self._x, self._y = point_bytes_to_coord(self.point)
        return self._x

    def y(self) -> int:
        """
        Returns the y-coordinate of the Ed25519 point.

        :return: The y-coordinate of the Ed25519 point.
        :rtype: int
        """

        if self._y is None:
            self._x, self._y = point_bytes_to_coord(self.point)
        return self._y

    def raw_encoded(self) -> bytes:
        """
        Returns the raw encoded point data.

        :return: The raw encoded point as bytes.
        :rtype: bytes
        """

        return self.point

    def raw_decoded(self) -> bytes:
        """
        Returns the decoded raw bytes representation of the point coordinates.

        :return: The raw decoded bytes of the point coordinates.
        :rtype: bytes
        """

        return int_encode(self.x()) + int_encode(self.y())

    def __add__(self, point: IPoint) -> IPoint:
        """
        Performs addition of this point object with another point object (`point`).

        :param point: The other point object to add to `self`.
        :type point: IPoint

        :return: The resulting point after addition.
        :rtype: IPoint
        """

        return self.__class__(
            point_add(self.point, point.underlying_object())
        )

    def __radd__(self, point: IPoint) -> IPoint:
        """
        Performs addition of this point object with another point object (`point`).

        :param point: The other point object to add to `self`.
        :type point: IPoint

        :return: The resulting point after addition.
        :rtype: IPoint
        """

        return self + point

    def __mul__(self, scalar: int) -> IPoint:
        """
        Performs scalar multiplication of this point object with an integer scalar.

        :param scalar: The scalar integer to multiply the point by.
        :type scalar: int

        :return: The resulting point after scalar multiplication.
        :rtype: IPoint
        """

        if self.is_generator:
            return self.__class__(
                point_scalar_mul_base(scalar)
            )
        return self.__class__(
            point_scalar_mul(scalar, self.point)
        )

    def __rmul__(self, scalar: int) -> IPoint:
        """
        Performs scalar multiplication of this point object with an integer scalar.

        :param scalar: The scalar integer to multiply the point by.
        :type scalar: int

        :return: The resulting point after scalar multiplication.
        :rtype: IPoint
        """

        return self * scalar



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Optional, List
)

from ..utils import normalize_derivation
from ..exceptions import DerivationError
from .iderivation import IDerivation


class CustomDerivation(IDerivation):
    """
    This class implements the Custom standard for hierarchical deterministic wallets.
    Custom defines a specific path structure for deriving keys from a master seed.

    .. note::
        This class inherits from the ``IDerivation`` class, thereby ensuring that all functions are accessible.
    """

    def __init__(
        self, path: Optional[str] = None, indexes: Optional[List[int]] = None
    ) -> None:
        """
        Initialize a CustomDerivation object with the given path and indexes.

        :param path: The derivation path as a string. Default is None.
        :type path: Optional[str]
        :param indexes: A list of indexes for the derivation path. Default is None.
        :type indexes: Optional[List[int]]

        :return: No return
        :rtype: NoneType
        """

        super(CustomDerivation, self).__init__(path, indexes)

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the derivation class.

        :return: The name of the derivation class.
        :rtype: str
        """

        return "Custom"

    def from_path(self, path: str) -> "CustomDerivation":
        """
        Sets the derivation path and indexes list from a given path string.

        :param path: The path string to set the derivation path and indexes list.
                     It should be in the format like "m/0'/0".
        :type path: str

        :return: The instance of CustomDerivation after setting the derivation path and indexes list.
        :rtype: CustomDerivation
        """

        if not isinstance(path, str):
            raise DerivationError("Bad path instance", expected=str, got=type(path))
        elif path[0:2] != "m/":
            raise DerivationError(
                f"Bad path format", expected="like this type of path \"m/0'/0\"", got=path
            )

        self._path, self._indexes, self._derivations = normalize_derivation(path=path)
        return self

    def from_indexes(self, indexes: List[int]) -> "CustomDerivation":
        """
        Sets the derivation path and indexes list from a list of indexes.

        :param indexes: The list of indexes to set the derivation path and indexes list.
        :type indexes: List[int]

        :return: The instance of CustomDerivation after setting the derivation path and indexes list.
        :rtype: CustomDerivation
        """

        if not isinstance(indexes, list):
            raise DerivationError("Bad indexes instance", expected=list, got=type(indexes))

        self._path, self._indexes, self._derivations = normalize_derivation(indexes=indexes)
        return self

    def from_index(self, index: int, hardened: bool = False) -> "CustomDerivation":
        """
        Appends an index to the derivation path and indexes list based on whether it's hardened.

        :param index: The index to append to the derivation path and indexes list.
        :type index: int
        :param hardened: Indicates if the index is hardened (True) or not (False), defaults to False.
        :type hardened: bool

        :return: The instance of CustomDerivation after appending the index.
        :rtype: CustomDerivation
        """

        if not isinstance(index, int):
            raise DerivationError("Bad index instance", expected=int, got=type(index))

        self._indexes.append(index + 0x80000000) if hardened else self._indexes.append(index)
        path = f"{index}'" if hardened else f"{index}"
        self.from_path(
            f"{self._path}{path}" if self._path == "m/" else f"{self._path}/{path}"
        )
        return self

    def clean(self) -> "CustomDerivation":
        """
        Resets the derivation path, indexes, and derivations to their default values.

        :return: The instance of CustomDerivation after cleaning.
        :rtype: CustomDerivation
        """

        self._path, self._indexes, self._derivations = normalize_derivation(
            path=None, indexes=None
        )
        return self



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Any
from ed25519_blake2b import SigningKey

from .....consts import SLIP10_ED25519_CONST
from ....iecc import (
    IPublicKey, IPrivateKey
)
from .public_key import SLIP10Ed25519Blake2bPublicKey


class SLIP10Ed25519Blake2bPrivateKey(IPrivateKey):

    signing_key: SigningKey

    def __init__(self, signing_key: SigningKey) -> None:
        """
        Initializes the class instance with a signing key.

        :param signing_key: The signing key object used for cryptographic operations.
        :type signing_key: SigningKey
        """

        self.signing_key = signing_key

    @staticmethod
    def name() -> str:
        """
        Get the name of the ecc class.

        :return: The name of the ecc class.
        :rtype: str
        """

        return "SLIP10-Ed25519-Blake2b"

    @classmethod
    def from_bytes(cls, private_key: bytes) -> IPrivateKey:
        """
        Creates an instance of the private key from its byte representation.

        :param private_key: The byte representation of the private key.
        :type private_key: bytes

        :return: An instance of the private key.
        :rtype: IPrivateKey
        """

        try:
            return cls(SigningKey(private_key))
        except ValueError as ex:
            raise ValueError("Invalid private key bytes") from ex

    @staticmethod
    def length() -> int:
        """
        Returns the length of the private key in bytes.

        :return: The length of the private key in bytes.
        :rtype: int
        """

        return SLIP10_ED25519_CONST.PRIVATE_KEY_BYTE_LENGTH

    def underlying_object(self) -> Any:
        """
        Returns the underlying object of the signing key.

        :return: The underlying object of the signing key.
        :rtype: Any
        """

        return self.signing_key

    def raw(self) -> bytes:
        """
        Returns the raw bytes representation of the signing key.

        :return: The raw bytes of the signing key.
        :rtype: bytes
        """

        return self.signing_key.to_bytes()

    def public_key(self) -> IPublicKey:
        """
        Returns the public key associated with this signing key.

        :return: The public key as an instance of IPublicKey.
        :rtype: IPublicKey
        """

        return SLIP10Ed25519Blake2bPublicKey(self.signing_key.get_verifying_key())



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import List

__name__: str = "hdwallet"
__version__: str = "v3.6.1"
__license__: str = "MIT"
__author__: str = "Meheret Tesfaye Batu"
__email__: str = "meherett.batu@gmail.com"
__documentation__: str = "https://hdwallet.readthedocs.com"
__description__: str = "Python-based library implementing a Hierarchical Deterministic (HD) Wallet generator for 200+ cryptocurrencies."
__url__: str = "https://hdwallet.io"
__source__: str = "https://github.com/hdwallet-io/python-hdwallet"
__changelog__: str = f"{__source__}/blob/master/CHANGELOG.md"
__tracker__: str = f"{__source__}/issues"
__keywords__: List[str] = [
    "ecc", "kholaw", "slip10", "ed25519", "nist256p1", "secp256k1",  # ECC keywords
    "hd", "bip32", "bip44", "bip49", "bip84", "bip86", "bip141", "monero", "cardano",  # HD keywords
    "entropy", "mnemonic", "seed", "bip39", "algorand", "electrum",  # Entropy, Mnemonic and Seed keywords
    "cryptocurrencies", "bitcoin", "ethereum", "cryptography", "cli", "cip1852"  # Other keywords
]
__websites__: List[str] = [
    "https://talonlab.org",
    "https://talonlab.gitbook.io/hdwallet",
    __documentation__,
    "https://hdwallet.online",
    "https://hd.wallet",  # On Web3 domain
    __url__
]



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    List, Union
)

import os
import re

from ..exceptions import EntropyError
from ..utils import (
    get_bytes, bytes_to_string, bytes_to_integer
)


class IEntropy:
    
    _entropy: str
    _strength: int

    strengths: List[int]

    def __init__(self, entropy: Union[bytes, str]) -> None:
        """
        Initialize the IEntropy class with the given entropy.

        :param entropy: The entropy value.
        :type entropy: Union[bytes, str]

        :return: No return
        :rtype: NoneType
        """

        try:
            strength: int = len(get_bytes(entropy))
            if self.name() == "Electrum-V2":
                if not self.are_entropy_bits_enough(get_bytes(entropy)):
                    raise EntropyError("Entropy bits are not enough")
                self._strength = bytes_to_integer(get_bytes(entropy)).bit_length()
            else:
                if not self.is_valid_bytes_strength(strength):
                    raise EntropyError("Unsupported entropy strength")
                self._strength = strength * 8
            self._entropy = bytes_to_string(entropy)
        except ValueError:
            raise EntropyError("Invalid entropy data")

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the entropy class.

        :return: The name of the entropy class.
        :rtype: str
        """

    def entropy(self) -> str:
        """
        Get entropy hex string.

        :return: The entropy value.
        :rtype: str
        """

        return self._entropy

    def strength(self) -> int:
        """
        :return: The strength of the entropy in bits.
        :rtype: int
        """

        return self._strength

    @classmethod
    def generate(cls, strength: int) -> str:
        """
        Generates a new entropy hex string with the given strength.

        :param strength: The entropy value.
        :type strength: int

        :return: The generated entropy hex string.
        :rtype: str
        """

        return bytes_to_string(
            os.urandom(strength // 8)
        )

    @classmethod
    def is_valid(cls, entropy: str) -> bool:
        """
        Checks if the given entropy is valid.

        :param entropy: Hex string representing entropy
        :type entropy: str

        :return: True if is valid, False otherwise.
        :rtype: bool
        """

        return isinstance(entropy, str) and bool(re.fullmatch(
            r'^[0-9a-fA-F]+$', entropy
        )) and cls.is_valid_strength(len(entropy) * 4)

    @classmethod
    def is_valid_strength(cls, strength: int) -> bool:
        """
        Checks if the given strength is valid.

        :param strength: The strength to check.
        :type strength: int

        :return: True if the strength is valid, False otherwise.
        :rtype: bool
        """

        return strength in cls.strengths

    @classmethod
    def is_valid_bytes_strength(cls, bytes_strength: int) -> bool:
        """
        Checks if the given byte strength is valid.

        :param bytes_strength: The byte strength to check.
        :type bytes_strength: int

        :return: True if the strength is valid, False otherwise.
        :rtype: bool
        """

        return cls.is_valid_strength(bytes_strength * 8)

    def are_entropy_bits_enough(self, entropy: Union[bytes, int]) -> bool:
        """
        Checks if the entropy bits are enough.

        :param entropy: The entropy value.
        :type entropy: Union[bytes, int]

        :return: True if the strength is valid, False otherwise.
        :rtype: bool
        """

        if self.name() != "Electrum-V2":
            raise NotImplemented



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from abc import (
    ABC, abstractmethod
)
from typing import (
    Union, Dict, List, Tuple, Optional
)

import os

from ..exceptions import MnemonicError
from ..entropies import IEntropy


class IMnemonic(ABC):

    _mnemonic: List[str]
    _words: int
    _language: str
    _mnemonic_type: Optional[str]

    words_list: List[int]
    languages: List[str]
    wordlist_path: Dict[str, str]

    def __init__(self, mnemonic: Union[str, List[str]], **kwargs) -> None:
        """
        Initialize an instance of IMnemonic with a mnemonic.

        :param mnemonic: The mnemonic to initialize with, which can be a string or a list of strings.
        :type mnemonic: Union[str, List[str]]
        :param kwargs: Additional keyword arguments.

        :return: No return
        :rtype: NoneType
        """

        self._mnemonic: List[str] = self.normalize(mnemonic)
        if not self.is_valid(self._mnemonic, **kwargs):
            raise MnemonicError("Invalid mnemonic words")

        _, self._language = self.find_language(self._mnemonic)
        self._mnemonic_type = kwargs.get("mnemonic_type", None)
        self._words = len(self._mnemonic)

    @classmethod
    def name(cls) -> str:
        pass

    def mnemonic(self) -> str:
        """
        Get the mnemonic as a single string.

        :return: The mnemonic as a single string joined by spaces.
        :rtype: str
        """

        return " ".join(self._mnemonic)

    def mnemonic_type(self) -> str:
        """
        Retrieves the type of the mnemonic.

        :return: The type of the mnemonic.
        :rtype: str
        """

        raise NotImplemented

    def language(self) -> str:
        """
        Get the formatted language value.

        :return: The formatted language string where each part is capitalized.
        :rtype: str
        """
        return self._language

    def words(self) -> int:
        """
        :return: The words.
        :rtype: int
        """

        return self._words

    @classmethod
    @abstractmethod
    def from_words(cls, words: int, language: str) -> str:
        pass

    @classmethod
    @abstractmethod
    def from_entropy(cls, entropy: Union[str, bytes, IEntropy], language: str, **kwargs) -> str:
        pass

    @classmethod
    @abstractmethod
    def encode(cls, entropy: Union[str, bytes], language: str) -> str:
        pass

    @classmethod
    @abstractmethod
    def decode(cls, mnemonic: Union[str, List[str]], **kwargs) -> str:
        pass

    @classmethod
    def get_words_list_by_language(
        cls, language: str, wordlist_path: Optional[Dict[str, str]] = None
    ) -> List[str]:
        """
        Retrieves the word list for the specified language.

        :param language: The language for which to get the word list.
        :type language: str
        :param wordlist_path: Optional dictionary mapping language names to file paths of their word lists.
        :type wordlist_path: Optional[Dict[str, str]]

        :return: A list of words for the specified language.
        :rtype: List[str]
        """

        wordlist_path = cls.wordlist_path if wordlist_path is None else wordlist_path
        with open(os.path.join(os.path.dirname(__file__), wordlist_path[language]), "r", encoding="utf-8") as fin:
            words_list: List[str] = [
                word.strip() for word in fin.readlines() if word.strip() != "" and not word.startswith("#")
            ]
        return words_list

    @classmethod
    def find_language(
        cls, mnemonic: List[str], wordlist_path: Optional[Dict[str, str]] = None
    ) -> Union[str, Tuple[List[str], str]]:
        """
        Finds the language of the given mnemonic by checking against available word lists.

        :param mnemonic: The mnemonic to check, represented as a list of words.
        :type mnemonic: List[str]
        :param wordlist_path: Optional dictionary mapping language names to file paths of their word lists.
        :type wordlist_path: Optional[Dict[str, str]]

        :return: A tuple containing the word list and the language name if found.
        :rtype: Union[str, Tuple[List[str], str]]
        """

        for language in cls.languages:
            try:
                words_list: list = cls.normalize(
                    cls.get_words_list_by_language(
                        language=language, wordlist_path=wordlist_path
                    )
                )
                words_list_with_index: dict = {
                    words_list[i]: i for i in range(len(words_list))
                }
                for word in mnemonic:
                    try:
                        words_list_with_index[word]
                    except KeyError as ex:
                        raise MnemonicError(f"Unable to find word {word}") from ex
                return words_list, language
            except (MnemonicError, ValueError):
                continue
        raise MnemonicError(f"Invalid language for mnemonic '{mnemonic}'")

    @classmethod
    def is_valid(cls, mnemonic: Union[str, List[str]], **kwargs) -> bool:
        """
        Checks if the given mnemonic is valid.

        :param mnemonic: The mnemonic to check.
        :type mnemonic: str
        :param kwargs: Additional keyword arguments.

        :return: True if the strength is valid, False otherwise.
        :rtype: bool
        """

        try:
            cls.decode(mnemonic=mnemonic, **kwargs)
            return True
        except (ValueError, MnemonicError):
            return False

    @classmethod
    def is_valid_language(cls, language: str) -> bool:
        """
        Checks if the given language is valid.

        :param language: The language to check.
        :type language: str

        :return: True if the strength is valid, False otherwise.
        :rtype: bool
        """

        return language in cls.languages

    @classmethod
    def is_valid_words(cls, words: int) -> bool:
        """
        Checks if the given words is valid.

        :param words: The words to check.
        :type words: int

        :return: True if the strength is valid, False otherwise.
        :rtype: bool
        """

        return words in cls.words_list

    @classmethod
    def normalize(cls, mnemonic: Union[str, List[str]]) -> List[str]:
        """
        Normalizes the given mnemonic by splitting it into a list of words if it is a string.

        :param mnemonic: The mnemonic value, which can be a single string of words or a list of words.
        :type mnemonic: Union[str, List[str]]

        :return: A list of words from the mnemonic.
        :rtype: List[str]
        """

        return mnemonic.split() if isinstance(mnemonic, str) else mnemonic



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from __future__ import annotations

from typing import Any
from abc import (
    ABC, abstractmethod
)

from .ipublic_key import IPublicKey


class IPrivateKey(ABC):

    @staticmethod
    @abstractmethod
    def name() -> str:
        """
        Get the name of the ecc class.

        :return: The name of the ecc class.
        :rtype: str
        """

    @classmethod
    @abstractmethod
    def from_bytes(cls, private_key: bytes) -> "IPrivateKey":
        """
        Create an IPrivateKey instance from a byte representation.

        :param private_key: The byte sequence representing the private key.
        :type private_key: bytes

        :return: A new IPrivateKey instance initialized from the byte representation.
        :rtype: IPrivateKey
        """

    @abstractmethod
    def raw(self) -> bytes:
        """
        Get the raw byte representation of the object.

        :return: The raw byte representation of the object.
        :rtype: bytes
        """

    @abstractmethod
    def public_key(self) -> IPublicKey:
        """
        Get the public key represented by this object.

        :return: The IPublicKey representation of the public key.
        :rtype: IPublicKey
        """

    @abstractmethod
    def underlying_object(self) -> Any:
        """
        Retrieve the underlying object represented by this instance.

        :return: The underlying object represented by this instance.
        :rtype: Any
        """

    @staticmethod
    @abstractmethod
    def length() -> int:
        """
        Get the length of the object.

        :return: The length of the object.
        :rtype: int
        """

    @classmethod
    def is_valid_bytes(cls, private_key: bytes) -> bool:
        """
        Checks if the given point is a valid bytes.

        :param private_key: The bytes to be validated.
        :type private_key: bytes

        :return: True if the point is valid, False otherwise.
        :rtype: bool
        """

        try:
            cls.from_bytes(private_key)
            return True
        except ValueError:
            return False



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Tuple, Union, Optional, Type
)

from ..eccs import (
    IEllipticCurveCryptography,
    SLIP10Secp256k1ECC,
    SLIP10Ed25519ECC,
    SLIP10Nist256p1ECC,
    KholawEd25519ECC,
    SLIP10Ed25519Blake2bECC,
    SLIP10Ed25519MoneroECC
)
from ..utils import (
    normalize_index, normalize_derivation, index_tuple_to_string
)
from ..exceptions import DerivationError
from .iderivation import IDerivation


class HDWDerivation(IDerivation):
    """
    This class implements the HDW standard for hierarchical deterministic wallets.
    HDW defines a specific path structure for deriving keys from a master seed.

    .. note::
        This class inherits from the ``IDerivation`` class, thereby ensuring that all functions are accessible.
    """

    _account: Union[Tuple[int, bool], Tuple[int, int, bool]]
    _ecc: Tuple[int, bool]
    _address: Union[Tuple[int, bool], Tuple[int, int, bool]]

    def __init__(
        self,
        account: Union[str, int, Tuple[int, int]] = 0,
        ecc: Optional[Union[str, int, Type[IEllipticCurveCryptography]]] = SLIP10Secp256k1ECC,
        address: Union[str, int, Tuple[int, int]] = 0
    ) -> None:
        """
        Initialize a HDW derivation path with specified parameters.

        :param account: The HDW account index or tuple. Defaults to 0.
        :type account: Union[str, int, Tuple[int, int]]
        :param ecc: The HDW ecc index. 
        :type ecc: Union[str, int, Type[IEllipticCurveCryptography]]
        :param address: The HDW address index or tuple. Defaults to 0.
        :type address: Union[str, int, Tuple[int, int]]

        :return: None
        """
        super(HDWDerivation, self).__init__()

        self._account = normalize_index(index=account, hardened=True)
        self._ecc = normalize_index(
            index=self.get_ecc_value(ecc=ecc, name_only=False), hardened=False
        )
        self._address = normalize_index(index=address, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._ecc)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the derivation class.

        :return: The name of the derivation class.
        :rtype: str
        """

        return "HDW"

    def get_ecc_value(
        self, ecc: Union[str, int, Type[IEllipticCurveCryptography]], name_only: bool = False
    ):

        if isinstance(ecc, IEllipticCurveCryptography):
            curve = ecc.NAME
        elif isinstance(ecc, type) and issubclass(ecc, IEllipticCurveCryptography):
            curve = ecc.NAME
        else:
            curve = ecc

        slip10_secp256k1 = [SLIP10Secp256k1ECC.NAME, 0, '0']
        slip10_ed25519 = [SLIP10Ed25519ECC.NAME, 1, '1']
        slip10_nist256p1 = [SLIP10Nist256p1ECC.NAME, 2, '2']
        kholaw_ed25519 = [KholawEd25519ECC.NAME, 3, '3']
        slip10_ed25519_blake2b = [SLIP10Ed25519Blake2bECC.NAME, 4, '4']
        slip10_ed25519_monero = [SLIP10Ed25519MoneroECC.NAME, 5, '5']

        expected_ecc = (
            slip10_secp256k1 + slip10_ed25519 + slip10_nist256p1 +
            kholaw_ed25519 + slip10_ed25519_blake2b + slip10_ed25519_monero
        )

        if curve not in expected_ecc:
            raise DerivationError(
            f"Bad {self.name()} ECC index",
                expected=expected_ecc, got=curve
            )

        if curve in slip10_secp256k1:
            return SLIP10Secp256k1ECC.NAME if name_only else 0
        if curve in slip10_ed25519:
            return SLIP10Ed25519ECC.NAME if name_only else 1
        if curve in slip10_nist256p1:
            return SLIP10Nist256p1ECC.NAME if name_only else 2
        if curve in kholaw_ed25519:
            return KholawEd25519ECC.NAME if name_only else 3
        if curve in slip10_ed25519_blake2b:
            return SLIP10Ed25519Blake2bECC.NAME if name_only else 4
        if curve in slip10_ed25519_monero:
            return SLIP10Ed25519MoneroECC.NAME if name_only else 5

    def from_account(self, account: Union[str, int, Tuple[int, int]]) -> "HDWDerivation":
        """
        Set the object's `_account` attribute to the specified account index or tuple,
        updating `_path`, `_indexes`, and `_derivations` accordingly.

        :param account: The account index or tuple to set. Can be a string, integer, or tuple of two integers.
        :type account: Union[str, int, Tuple[int, int]]

        :return: The updated `HDWDerivation` object itself after setting the account.
        :rtype: HDWDerivation
        """

        self._account = normalize_index(index=account, hardened=True)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._ecc)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def from_ecc(self, ecc: Union[str, int, Type[IEllipticCurveCryptography]]) -> "HDWDerivation":
        """
        Set the object's `_ecc` attribute to the specified ecc index or key,
        updating `_path`, `_indexes`, and `_derivations` accordingly.

        :param ecc: The ecc index or key to set. Can be a string, integer, or one of the predefined keys.
        :type ecc: Union[str, int, Type[IEllipticCurveCryptography]]

        :return: The updated `HDWDerivation` object itself after setting the ecc.
        :rtype: HDWDerivation
        """

        self._ecc = normalize_index(
            index=self.get_ecc_value(ecc=ecc, name_only=False), hardened=False
        )
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._ecc)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def from_address(self, address: Union[str, int, Tuple[int, int]]) -> "HDWDerivation":
        """
        Set the object's `_address` attribute to the specified address index or tuple of indexes,
        updating `_path`, `_indexes`, and `_derivations` accordingly.

        :param address: The address index or tuple of indexes to set. Should be non-hardened.
        :type address: Union[str, int, Tuple[int, int]]

        :return: The updated `HDWDerivation` object itself after setting the address.
        :rtype: HDWDerivation
        """

        self._address = normalize_index(index=address, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._ecc)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def clean(self) -> "HDWDerivation":
        """
        Reset the object's attributes related to HDW derivation to their initial states or defaults.

        :return: The updated `HDWDerivation` object itself after cleaning.
        :rtype: HDWDerivation
        """

        self._account = normalize_index(index=0, hardened=True)
        self._address = normalize_index(index=0, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._ecc)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def account(self) -> int:
        """
        Retrieve the account value from the object's `_account` attribute.

        Checks the length of `_account`. If it equals 3, returns the second
        element; otherwise, returns the first element.

        :return: The account value stored in `_account`.
        :rtype: int
        """

        return (
            self._account[1] if len(self._account) == 3 else self._account[0]
        )

    def ecc(self, name_only: bool = True) -> str:
        """
        Retrieve the ecc value from the object's eccs dictionary.

        Iterates through the `eccs` dictionary, and if a value matches the first element of `_ecc`,
        sets the corresponding key as the ecc value.

        :param name_only: Return the ECC name if True, or its index if False.
        :type name_only: bool
        :return: The key from the `eccs` dictionary that corresponds to the `_ecc` value, or `None` if not found.
        :rtype: str
        """
        return self.get_ecc_value(ecc=self._ecc[0], name_only=name_only)

    def address(self) -> int:
        """
        Retrieve the address from the object.

        :return: The address value.
        :rtype: int
        """

        return (
            self._address[1] if len(self._address) == 3 else self._address[0]
        )



#!/usr/bin/env python3

# Copyright © 2020-2025 , Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Optional, Union, List
)

import cbor2
import re

from ..mnemonics import (
    IMnemonic, BIP39Mnemonic
)
from ..cryptocurrencies import Cardano
from ..crypto import blake2b_256
from ..exceptions import (
    Error, MnemonicError, SeedError
)
from ..utils import (
    get_bytes, bytes_to_string
)
from . import (
    ISeed, BIP39Seed
)


class CardanoSeed(ISeed):
    """
    This class generates a root extended private key from a given seed using the
    cardano standard. The cardano standard defines a method for generating mnemonic
    phrases and converting them into a binary seed used for hierarchical
    deterministic wallets.

    .. note::
        This class inherits from the ``ISeed`` class, thereby ensuring that all functions are accessible.
    """

    _cardano_type: str
    lengths: List[int] = [
        32,  # Byron-Icarus and Shelly-Icarus
        128,  # Byron-Ledger and Shelly-Ledger
        64  # Byron-Legacy
    ]

    def __init__(
        self, seed: str, cardano_type: str = Cardano.TYPES.BYRON_ICARUS, passphrase: Optional[str] = None
    ) -> None:
        """
        Initialize a CardanoSeed object.

        :param seed: The seed value used to generate Cardano keys.
        :type seed: str
        :param cardano_type: The type of Cardano seed. Defaults to Cardano.TYPES.BYRON_ICARUS.
        :type cardano_type: str, optional

        :param passphrase: Optional passphrase for deriving the seed. Defaults to None.
        :type passphrase: str, optional
        """

        super(CardanoSeed, self).__init__(
            seed=seed, cardano_type=cardano_type, passphrase=passphrase
        )

        if cardano_type not in Cardano.TYPES.get_cardano_types():
            raise SeedError(
                "Invalid Cardano type", expected=Cardano.TYPES.get_cardano_types(), got=cardano_type
            )

        self._cardano_type = cardano_type
        self._seed = seed

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the seeds class.

        :return: The name of the seeds class.
        :rtype: str
        """

        return "Cardano"

    def cardano_type(self) -> str:
        """
        Returns the Cardano type associated with this instance.

        :return: The Cardano type as a string.
        :rtype: str
        """

        return self._cardano_type

    @classmethod
    def is_valid(cls, seed: str, cardano_type: str = Cardano.TYPES.BYRON_ICARUS) -> bool:
        """
        Checks if the given seed is valid.

        :param seed: Hex string representing seed
        :type seed: str
        :param cardano_type: The type of Cardano seed. Defaults to Cardano.TYPES.BYRON_ICARUS.
        :type cardano_type: str, optional

        :return: True if is valid, False otherwise.
        :rtype: bool
        """

        if not isinstance(seed, str) or not bool(re.fullmatch(
            r'^[0-9a-fA-F]+$', seed
        )): return False

        if cardano_type in [Cardano.TYPES.BYRON_ICARUS, Cardano.TYPES.SHELLEY_ICARUS]:
            return len(seed) == cls.lengths[0]
        elif cardano_type in [Cardano.TYPES.BYRON_LEDGER, Cardano.TYPES.SHELLEY_LEDGER]:
            return len(seed) == cls.lengths[1]
        elif cardano_type == Cardano.TYPES.BYRON_LEGACY:
            return len(seed) == cls.lengths[2]
        else:
            raise SeedError(
                "Invalid Cardano type", expected=Cardano.TYPES.get_cardano_types(), got=cardano_type
            )

    @classmethod
    def from_mnemonic(
        cls,
        mnemonic: Union[str, IMnemonic],
        passphrase: Optional[str] = None,
        cardano_type: str = Cardano.TYPES.BYRON_ICARUS
    ) -> str:
        """
        Generates a Cardano wallet seed based on the specified cardano_type.

        This method allows generating different types of Cardano wallet seeds based on the `cardano_type` parameter.

        :param mnemonic: The mnemonic phrase to be used for seed generation. Can be a string or an instance of `IMnemonic`.
        :type mnemonic: Union[str, IMnemonic]

        :param passphrase: Optional passphrase used for seed derivation, default is `None`.
        :type passphrase: Optional[str]

        :param cardano_type: Specifies the type of Cardano seed to generate. Default is Byron Icarus.
                             Valid options are defined in `Cardano.TYPES`.
        :type cardano_type: str

        :return: The generated Cardano wallet seed as a string.
        :rtype: str
        """

        if cardano_type == Cardano.TYPES.BYRON_ICARUS:
            return cls.generate_byron_icarus(mnemonic=mnemonic)
        if cardano_type == Cardano.TYPES.BYRON_LEDGER:
            return cls.generate_byron_ledger(
                mnemonic=mnemonic, passphrase=passphrase
            )
        if cardano_type == Cardano.TYPES.BYRON_LEGACY:
            return cls.generate_byron_legacy(mnemonic=mnemonic)
        if cardano_type == Cardano.TYPES.SHELLEY_ICARUS:
            return cls.generate_shelley_icarus(mnemonic=mnemonic)
        elif cardano_type == Cardano.TYPES.SHELLEY_LEDGER:
            return cls.generate_shelley_ledger(
                mnemonic=mnemonic, passphrase=passphrase
            )
        raise Error(
            "Invalid Cardano type", expected=Cardano.TYPES.get_cardano_types(), got=cardano_type
        )

    @classmethod
    def generate_byron_icarus(cls, mnemonic: Union[str, IMnemonic]) -> str:
        """
        Generates a Byron Icarus seed from a given mnemonic phrase.

        :param mnemonic: The mnemonic phrase to be used for seed generation. Can be a string or an instance of `IMnemonic`.
        :type mnemonic: Union[str, IMnemonic]

        :return: The derived Byron Icarus seed as a string.
        :rtype: str
        """

        mnemonic = (
            mnemonic.mnemonic()
            if isinstance(mnemonic, IMnemonic) else
            mnemonic
        )
        if not BIP39Mnemonic.is_valid(mnemonic=mnemonic):
            raise MnemonicError(f"Invalid {BIP39Mnemonic.name()} mnemonic words")

        return BIP39Mnemonic.decode(mnemonic=mnemonic)

    @classmethod
    def generate_byron_ledger(cls, mnemonic: Union[str, IMnemonic], passphrase: Optional[str] = None) -> str:
        """
        Generates a Byron Ledger seed from a given mnemonic phrase and optional passphrase.

        :param mnemonic: The mnemonic phrase to be used for seed generation. Can be a string or an instance of `IMnemonic`.
        :type mnemonic: Union[str, IMnemonic]

        :param passphrase: An optional passphrase to be used in the seed generation process.
        :type passphrase: Optional[str]

        :return: The derived Byron Ledger seed as a string.
        :rtype: str
        """

        mnemonic = (
            mnemonic.mnemonic()
            if isinstance(mnemonic, IMnemonic) else
            mnemonic
        )
        return BIP39Seed.from_mnemonic(mnemonic=mnemonic, passphrase=passphrase)

    @classmethod
    def generate_byron_legacy(cls, mnemonic: Union[str, IMnemonic]) -> str:
        """
        Generates a Byron Legacy seed from a given mnemonic phrase.

        :param mnemonic: The mnemonic phrase to be used for seed generation. Can be a string or an instance of `IMnemonic`.
        :type mnemonic: Union[str, IMnemonic]

        :return: The derived Byron Legacy seed as a string.
        :rtype: str
        """

        mnemonic = (
            mnemonic.mnemonic()
            if isinstance(mnemonic, IMnemonic) else
            mnemonic
        )
        if not BIP39Mnemonic.is_valid(mnemonic=mnemonic):
            raise MnemonicError(f"Invalid {BIP39Mnemonic.name()} mnemonic words")

        return bytes_to_string(blake2b_256(
            cbor2.dumps(get_bytes(BIP39Mnemonic.decode(mnemonic=mnemonic)))
        ))

    @classmethod
    def generate_shelley_icarus(cls, mnemonic: Union[str, IMnemonic]) -> str:
        """
        Generates a Shelley Icarus seed from a given mnemonic phrase.

        :param mnemonic: The mnemonic phrase to be used for seed generation. Can be a string or an instance of `IMnemonic`.
        :type mnemonic: Union[str, IMnemonic]

        :return: The derived Shelley Icarus seed as a string.
        :rtype: str
        """

        return cls.generate_byron_icarus(
            mnemonic=mnemonic
        )

    @classmethod
    def generate_shelley_ledger(cls, mnemonic: str, passphrase: Optional[str] = None) -> str:
        """
        Generates a Shelley ledger seed from a given mnemonic phrase and optional passphrase.

        :param mnemonic: The mnemonic phrase to be used for seed generation.
        :type mnemonic: str
        :param passphrase: An optional passphrase to strengthen the security of the seed. Defaults to None.
        :type passphrase: Optional[str]

        :return: The derived Shelley ledger seed as a string.
        :rtype: str
        """

        return cls.generate_byron_ledger(
            mnemonic=mnemonic, passphrase=passphrase
        )



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Optional, Union
)

import hmac
import hashlib

from ..eccs import KholawEd25519ECC
from ..addresses import AlgorandAddress
from ..cryptocurrencies import Algorand
from ..seeds import ISeed
from ..exceptions import (
    Error, SeedError, DerivationError
)
from ..utils import (
    get_bytes, bytes_to_integer, integer_to_bytes
)
from .bip32 import BIP32HD


class AlgorandHD(BIP32HD):

    def __init__(self) -> None:
        """
        Initialize a AlgorandHD instance.
        """

        super(AlgorandHD, self).__init__(
            ecc=KholawEd25519ECC
        )

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the Algorand class.

        :return: The name of the Algorand class.
        :rtype: str
        """
        return "Algorand"

    def from_seed(self, seed: Union[bytes, str, ISeed], **kwargs) -> "AlgorandHD":
        """
        Initializes the AlgorandHD instance from the given seed.

        :param seed: The seed to initialize the instance. It can be of type `bytes`, `str`, or `ISeed`.
        :type seed: Union[bytes, str, ISeed]
        :param kwargs: Additional keyword arguments.

        :return: The initialized AlgorandHD instance.
        :rtype: AlgorandHD
        """

        try:
            self._seed = get_bytes(
                seed.seed() if isinstance(seed, ISeed) else seed
            )
        except ValueError as error:
            raise SeedError("Invalid seed data")

        if len(self._seed) < 16:
            raise Error(f"Invalid seed length", expected="< 16", got=len(self._seed))

        def clamp_kL(kL: bytearray):
            kL[0] &= 0b11111000
            kL[31] &= 0b01111111
            kL[31] |= 0b01000000
            return kL

        k = hashlib.sha512(self._seed).digest()
        kL = bytearray(k[0:32])
        kR = k[32:64]

        while (kL[31] & 0b00100000) != 0:
            k = hmac.new(kL, kR, hashlib.sha512).digest()
            kL = bytearray(k[0:32])
            kR = k[32:64]

        kL = clamp_kL(kL)

        chain_code_bytes = hashlib.sha256(bytes([0x01]) + self._seed).digest()

        self._root_private_key = self._ecc.PRIVATE_KEY.from_bytes(bytes(kL + kR))
        self._root_chain_code = chain_code_bytes

        self._private_key, self._chain_code, self._parent_fingerprint = (
            self._root_private_key, self._root_chain_code, (integer_to_bytes(0x00) * 4)
        )
        self._root_public_key = self._root_private_key.public_key()
        self._public_key = self._root_public_key
        self._strict = True
        self.__update__()
        return self

    def drive(self, index: int) -> Optional["AlgorandHD"]:
        """
        Drives the AlgorandHD instance forward along the derivation path by deriving a child key at the given index.

        :param index: The index to derive the child key.
        :type index: int

        :return: The updated AlgorandHD instance with the derived child key, or None if the derivation fails.
        :rtype: Optional[AlgorandHD]
        """

        G = 9  # Peikert's suggestion for 8 levels of derivation security
        index_bytes = integer_to_bytes(index, 4, endianness="little")
        cc = self._chain_code

        if cc is None:
            raise DerivationError("Chain code is not set")

        # Hardened derivation requires private key
        if index & 0x80000000:
            if self._private_key is None:
                raise DerivationError("Private key required for hardened derivation")

            kL = self._private_key.raw()[:32]
            kR = self._private_key.raw()[32:]

            data = bytes([0x00]) + kL + kR + index_bytes
            z = hmac.new(cc, data, hashlib.sha512).digest()
            data = bytes([0x01]) + kL + kR + index_bytes
            child_cc = hmac.new(cc, data, hashlib.sha512).digest()[32:]

            zL, zR = z[:32], z[32:]

            def trunc_256_minus_g_bits(buf: bytes, g: int) -> bytes:
                if g < 0 or g > 256:
                    raise ValueError("g must be between 0 and 256")
                out = bytearray(buf)
                remaining = g
                for i in range(len(out) - 1, -1, -1):
                    if remaining >= 8:
                        out[i] = 0
                        remaining -= 8
                    elif remaining > 0:
                        out[i] &= (0xFF >> remaining)
                        break
                return bytes(out)

            truncated_zL = trunc_256_minus_g_bits(zL, G)
            kL_int = bytes_to_integer(kL, endianness="little")
            zL_int = bytes_to_integer(truncated_zL, endianness="little")
            child_kL_int = kL_int + (8 * zL_int)

            if child_kL_int >= 2 ** 255:
                raise DerivationError("zL * 8 + kL exceeds Ed25519 scalar limit")

            child_kL = integer_to_bytes(child_kL_int, 32, endianness="little")
            child_kR = integer_to_bytes(
                (bytes_to_integer(kR, "little") + bytes_to_integer(zR, "little")) % (2 ** 256),
                32, "little"
            )

            child_key = self._ecc.PRIVATE_KEY.from_bytes(child_kL + child_kR)
            self._private_key = child_key
            self._parent_fingerprint = get_bytes(self.fingerprint())
            self._public_key = child_key.public_key()

        # Non-hardened derivation (public key supported)
        else:
            if self._public_key is None:
                raise DerivationError("Public key is required for non-hardened derivation")

            A = self._public_key.raw_compressed()[1:]
            data = bytes([0x02]) + A + index_bytes
            z = hmac.new(cc, data, hashlib.sha512).digest()
            data = bytes([0x03]) + A + index_bytes
            child_cc = hmac.new(cc, data, hashlib.sha512).digest()[32:]

            zL = z[:32]

            def trunc_256_minus_g_bits(buf: bytes, g: int) -> bytes:
                if g < 0 or g > 256:
                    raise ValueError("g must be between 0 and 256")
                out = bytearray(buf)
                remaining = g
                for i in range(len(out) - 1, -1, -1):
                    if remaining >= 8:
                        out[i] = 0
                        remaining -= 8
                    elif remaining > 0:
                        out[i] &= (0xFF >> remaining)
                        break
                return bytes(out)

            truncated_zL = trunc_256_minus_g_bits(zL, G)
            scalar = 8 * bytes_to_integer(truncated_zL, "little")
            new_point = self._public_key.point() + (self._ecc.GENERATOR * scalar)
            self._parent_fingerprint = get_bytes(self.fingerprint())
            self._public_key = self._ecc.PUBLIC_KEY.from_point(new_point)

        self._chain_code = child_cc
        self._depth += 1
        self._index = index
        self._fingerprint = get_bytes(self.fingerprint())

        return self

    def root_xprivate_key(
        self, version: Union[bytes, int] = Algorand.NETWORKS.MAINNET.XPRIVATE_KEY_VERSIONS.P2PKH, encoded: bool = True
    ) -> Optional[str]:
        """
        Generates the root extended private key (xprv) in serialized format.

        :param version: The version bytes for the extended key. Defaults to Algorand mainnet P2PKH version.
        :type version: Union[bytes, int]
        :param encoded: Whether to return the key in encoded format. Defaults to True.
        :type encoded: bool

        :return: The root extended private key (xprv) in serialized format, or None if the chain code is not set.
        :rtype: Optional[str]
        """
        return super(AlgorandHD, self).root_xprivate_key(version=version, encoded=encoded)

    def xprivate_key(
        self, version: Union[bytes, int] = Algorand.NETWORKS.MAINNET.XPRIVATE_KEY_VERSIONS.P2PKH, encoded: bool = True
    ) -> Optional[str]:
        """
        Retrieves the extended private key (xprivate key) as a serialized string.

        :param version: The version bytes or integer version of the xprivate key.
        :type version: Union[bytes, int]
        :param encoded: Flag indicating whether the key should be encoded.
        :type encoded: bool

        :return: The serialized xprivate key as a string, or None if the private key or chain code is not set.
        :rtype: Optional[str]
        """
        return super(AlgorandHD, self).xprivate_key(version=version, encoded=encoded)
    
    def address(self, **kwargs) -> str:
        """
        Generates a Algorand address using the AlgorandAddress encoding scheme.

        :param kwargs: Additional keyword arguments for encoding.

        :return: Algorand address encoded.
        :rtype: str
        """

        return AlgorandAddress.encode(public_key=self._public_key)



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    List, Dict, Type
)

from ..exceptions import EntropyError
from .algorand import (
    AlgorandEntropy, ALGORAND_ENTROPY_STRENGTHS
)
from .bip39 import (
    BIP39Entropy, BIP39_ENTROPY_STRENGTHS
)
from .electrum import (
    ElectrumV1Entropy, ELECTRUM_V1_ENTROPY_STRENGTHS,
    ElectrumV2Entropy, ELECTRUM_V2_ENTROPY_STRENGTHS
)
from .monero import (
    MoneroEntropy, MONERO_ENTROPY_STRENGTHS
)
from .ientropy import IEntropy


class ENTROPIES:
    """
    A class that manages a dictionary of entropy classes.

    This class provides methods to retrieve names and classes of various entropy implementations,
    as well as methods to validate and access specific entropy classes by name.

    Here are available entropy names and classes:

    +--------------+-------------------------------------------------------------+
    | Name         | Class                                                       |
    +==============+=============================================================+
    | Algorand     |  :class:`hdwallet.entropies.algorand.AlgorandEntropy`       |
    +--------------+-------------------------------------------------------------+
    | BIP39        |  :class:`hdwallet.entropies.bip39.BIP39Entropy`             |
    +--------------+-------------------------------------------------------------+
    | Electrum-V1  |  :class:`hdwallet.entropies.electrum.v1.ElectrumV1Entropy`  |
    +--------------+-------------------------------------------------------------+
    | Electrum-V2  | :class:`hdwallet.entropies.electrum.v2.ElectrumV2Entropy`   |
    +--------------+-------------------------------------------------------------+
    | Monero       | :class:`hdwallet.entropies.monero.MoneroEntropy`            |
    +--------------+-------------------------------------------------------------+
    """

    dictionary: Dict[str, Type[IEntropy]] = {
        AlgorandEntropy.name(): AlgorandEntropy,
        BIP39Entropy.name(): BIP39Entropy,
        ElectrumV1Entropy.name(): ElectrumV1Entropy,
        ElectrumV2Entropy.name(): ElectrumV2Entropy,
        MoneroEntropy.name(): MoneroEntropy
    }

    @classmethod
    def names(cls) -> List[str]:
        """
        Get the list of entropy class names.

        :return: A list of entropy class names.
        :rtype: List[str]
        """

        return list(cls.dictionary.keys())

    @classmethod
    def classes(cls) -> List[Type[IEntropy]]:
        """
        Get the list of entropy classes.

        :return: A list of entropy classes.
        :rtype: List[Type[IEntropy]]
        """

        return list(cls.dictionary.values())

    @classmethod
    def entropy(cls, name: str) -> Type[IEntropy]:
        """
        Retrieve an entropy class by its name.

        :param name: The name of the entropy class.
        :type name: str

        :return: The entropy class corresponding to the given name.
        :rtype: Type[IEntropy]
        """

        if not cls.is_entropy(name=name):
            raise EntropyError(
                "Invalid entropy name", expected=cls.names(), got=name
            )
        return cls.dictionary[name]

    @classmethod
    def is_entropy(cls, name: str) -> bool:
        """
        Check if a given name is a valid entropy name.

        :param name: The name to check.
        :type name: str

        :return: True if the name is valid, False otherwise.
        :rtype: bool
        """

        return name in cls.names()


__all__: List[str] = [
    "IEntropy",
    "ALGORAND_ENTROPY_STRENGTHS",
    "BIP39_ENTROPY_STRENGTHS",
    "ELECTRUM_V1_ENTROPY_STRENGTHS",
    "ELECTRUM_V2_ENTROPY_STRENGTHS",
    "MONERO_ENTROPY_STRENGTHS",
    "ENTROPIES"
] + [
    cls.__name__ for cls in ENTROPIES.classes()
]



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from tabulate import tabulate

import click

from ...cryptocurrencies import CRYPTOCURRENCIES


def list_cryptocurrencies():

    documents, table, headers = [], [], [
        "Cryptocurrency", "Symbol", "Coin Type", "Networks", "ECC"
    ]

    for cryptocurrency in CRYPTOCURRENCIES.classes():

        document: dict = {
            "name": cryptocurrency.NAME,
            "symbol": cryptocurrency.SYMBOL,
            "coin_type": cryptocurrency.COIN_TYPE,
            "networks": ", ".join(cryptocurrency.NETWORKS.get_networks()),
            "ecc": cryptocurrency.ECC.NAME
        }
        documents.append(document)

    for document in documents:
        table.append([
            document["name"],
            document["symbol"],
            document["coin_type"],
            document["networks"],
            document["ecc"],
        ])

    click.echo(tabulate(
        table,
        headers,
        colalign=(
            "left", "center", "center", "center", "center"
        )
    ))



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import List

from ..slip44 import CoinTypes
from ..eccs import KholawEd25519ECC
from ..consts import (
    Info, NestedNamespace, Entropies, Mnemonics, Seeds, HDs, Addresses, AddressTypes, Networks, Params, XPrivateKeyVersions, XPublicKeyVersions
)
from .icryptocurrency import (
    ICryptocurrency, INetwork
)


class Types(NestedNamespace):

    BYRON_ICARUS: str
    BYRON_LEDGER: str
    BYRON_LEGACY: str
    SHELLEY_ICARUS: str
    SHELLEY_LEDGER: str

    def get_cardano_types(self) -> List[str]:
        return list(self.__dict__.values())

    def is_cardano_type(self, cardano_type) -> bool:
        return cardano_type in self.get_cardano_types()


class Mainnet(INetwork):

    NAME = "mainnet"
    TYPE = 1
    PAYMENT_ADDRESS_HRP = "addr"
    REWARD_ADDRESS_HRP = "stake"
    XPRIVATE_KEY_VERSIONS = XPrivateKeyVersions({
        "P2PKH": 0x0f4331d4
    })
    XPUBLIC_KEY_VERSIONS = XPublicKeyVersions({
        "P2PKH": 0x0488b21e
    })


class Testnet(INetwork):

    NAME = "testnet"
    TYPE = 0
    PAYMENT_ADDRESS_HRP = "addr_test"
    REWARD_ADDRESS_HRP = "stake_test"
    XPRIVATE_KEY_VERSIONS = XPrivateKeyVersions({
        "P2PKH": 0x04358394
    })
    XPUBLIC_KEY_VERSIONS = XPublicKeyVersions({
        "P2PKH": 0x043587cf
    })


class Cardano(ICryptocurrency):

    NAME = "Cardano"
    SYMBOL = "ADA"
    INFO = Info({
        "SOURCE_CODE": "https://cardanoupdates.com",
        "WHITEPAPER": "https://docs.cardano.org/en/latest",
        "WEBSITES": [
            "https://www.cardano.org"
        ]
    })
    ECC = KholawEd25519ECC
    COIN_TYPE = CoinTypes.Cardano
    SUPPORT_BIP38 = False
    NETWORKS = Networks({
        "MAINNET": Mainnet, "TESTNET": Testnet
    })
    DEFAULT_NETWORK = NETWORKS.MAINNET
    ENTROPIES = Entropies({
        "BIP39"
    })
    MNEMONICS = Mnemonics({
        "BIP39"
    })
    SEEDS = Seeds({
        "CARDANO": "Cardano"
    })
    HDS = HDs({
        "CARDANO": "Cardano"
    })
    DEFAULT_HD = HDS.CARDANO
    DEFAULT_PATH = f"m/44'/{COIN_TYPE}'/0'/0/0"
    TYPES = Types({
        "BYRON_ICARUS": "byron-icarus",
        "BYRON_LEDGER": "byron-ledger",
        "BYRON_LEGACY": "byron-legacy",
        "SHELLEY_ICARUS": "shelley-icarus",
        "SHELLEY_LEDGER": "shelley-ledger"
    })
    ADDRESSES = Addresses({
        "CARDANO": "Cardano"
    })
    DEFAULT_ADDRESS = ADDRESSES.CARDANO
    SEMANTICS = [
        "p2pkh"
    ]
    DEFAULT_SEMANTIC = "p2pkh"
    ADDRESS_TYPES = AddressTypes({
        "PUBLIC_KEY": "public-key",
        "REDEMPTION": "redemption",
        "PAYMENT": "payment",
        "STAKING": "staking",
        "REWARD": "reward"
    })
    PARAMS = Params({
        "PUBLIC_KEY_ADDRESS": 0,
        "REDEMPTION_ADDRESS": 2,
        "PAYMENT_PREFIX": 0x00,
        "REWARD_PREFIX": 0x0e
    })



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Any
from ecdsa import SigningKey
from ecdsa import (
    curves, keys
)

from ....consts import SLIP10_SECP256K1_CONST
from ...iecc import (
    IPublicKey, IPrivateKey
)
from .public_key import SLIP10Nist256p1PublicKey


class SLIP10Nist256p1PrivateKey(IPrivateKey):
    
    signing_key: SigningKey

    def __init__(self, signing_key: SigningKey) -> None:
        """
        Initialize the private key object.

        :param signing_key: The signing key object.
        :type signing_key: SigningKey
        """

        self.signing_key = signing_key

    @staticmethod
    def name() -> str:
        """
        Get the name of the ecc class.

        :return: The name of the ecc class.
        :rtype: str
        """

        return "SLIP10-Nist256p1"

    @classmethod
    def from_bytes(cls, private_key: bytes) -> IPrivateKey:
        """
        Create a private key object from a byte string.

        :param private_key: The byte string representing the private key.
        :type private_key: bytes

        :return: An instance of the private key.
        :rtype: IPrivateKey
        """

        try:
            return cls(
                SigningKey.from_string(
                    private_key, curve=curves.NIST256p
                )
            )
        except keys.MalformedPointError as ex:
            raise ValueError("Invalid private key bytes") from ex

    @staticmethod
    def length() -> int:
        """
        Get the length of the private key in bytes.

        :return: The private key length in bytes.
        :rtype: int
        """

        return SLIP10_SECP256K1_CONST.PRIVATE_KEY_BYTE_LENGTH

    def underlying_object(self) -> Any:
        """
        Retrieve the underlying signing key object.

        :return: The underlying signing key object.
        :rtype: Any
        """

        return self.signing_key

    def raw(self) -> bytes:
        """
        Retrieve the raw byte representation of the signing key.

        :return: The raw bytes of the signing key.
        :rtype: bytes
        """

        return self.signing_key.to_string()

    def public_key(self) -> IPublicKey:
        """
        Retrieve the corresponding public key for the signing key.

        :return: The public key associated with the signing key.
        :rtype: IPublicKey
        """

        return SLIP10Nist256p1PublicKey(self.signing_key.get_verifying_key())



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..eccs import (
    IPublicKey, SLIP10Ed25519PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Sui
from ..crypto import blake2b_256
from ..utils import (
    bytes_to_string, integer_to_bytes
)
from .iaddress import IAddress


class SuiAddress(IAddress):

    key_type: bytes = integer_to_bytes(Sui.PARAMS.KEY_TYPE)
    address_prefix: str = Sui.PARAMS.ADDRESS_PREFIX

    @staticmethod
    def name() -> str:
        """
        Return the name "Sui".

        :return: The string "Sui".
        :rtype: str
        """

        return "Sui"

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encode a public key into a string format using a specific hashing algorithm.

        :param public_key: The public key to encode, can be bytes, string, or an IPublicKey object.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any

        :return: The encoded address as a string.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Ed25519PublicKey
        )
        return cls.address_prefix + bytes_to_string(blake2b_256(
            cls.key_type + public_key.raw_compressed()[1:]
        ))

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decode an address string and validate its format.

        :param address: The address string to decode.
        :type address: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any

        :return: The decoded address string.
        :rtype: str
        """

        address_prefix_got: str = address[:len(cls.address_prefix)]
        if cls.address_prefix != address_prefix_got:
            raise ValueError(f"Invalid address_prefix (expected: {cls.address_prefix}, got: {address_prefix_got})")
        address_no_prefix = address[len(cls.address_prefix):]

        if len(address_no_prefix) != 64:
            raise ValueError(f"Invalid length (expected: {64}, got: {len(address_no_prefix)})")

        return address_no_prefix



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.base32 import (
    encode_no_padding, decode
)
from ..eccs import (
    IPublicKey, SLIP10Secp256k1PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Filecoin
from ..crypto import (
    blake2b_32, blake2b_160
)
from ..utils import (
    get_bytes, integer_to_bytes, bytes_to_string
)
from ..exceptions import AddressError
from .iaddress import IAddress


class FilecoinAddress(IAddress):

    alphabet: str = Filecoin.PARAMS.ALPHABET
    address_prefix: str = Filecoin.PARAMS.ADDRESS_PREFIX
    address_types: dict = {
        "secp256k1": Filecoin.PARAMS.ADDRESS_TYPES.SECP256K1,
        "bls": Filecoin.PARAMS.ADDRESS_TYPES.BLS
    }

    @staticmethod
    def name() -> str:
        """
        Returns the name of the blockchain.

        :return: The name of the blockchain.
        :rtype: str
        """

        return "Filecoin"

    @classmethod
    def compute_checksum(cls, public_key_hash: bytes, address_type: int) -> bytes:
        return blake2b_32(
            integer_to_bytes(address_type) + public_key_hash
        )

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encodes the given public key into a Filecoin address.

        :param public_key: The public key to be encoded.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments, including "address_type".
        :type kwargs: Any

        :return: The encoded Filecoin address.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Secp256k1PublicKey
        )
        public_key_hash: bytes = blake2b_160(
            public_key.raw_uncompressed()
        )

        if not kwargs.get("address_type"):
            address_type: int = cls.address_types[Filecoin.DEFAULT_ADDRESS_TYPE]
        else:
            if kwargs.get("address_type") not in Filecoin.ADDRESS_TYPES.get_address_types():
                raise AddressError(
                    f"Invalid {cls.name()} address type",
                    expected=Filecoin.ADDRESS_TYPES.get_address_types(),
                    got=kwargs.get("address_type")
                )
            address_type: int = cls.address_types[kwargs.get("address_type")]

        checksum: bytes = cls.compute_checksum(public_key_hash, address_type)
        base32_encode: str = encode_no_padding(
            (public_key_hash + checksum).hex(), cls.alphabet
        )
        return (
            cls.address_prefix + chr(address_type + ord("0")) + base32_encode
        )

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decodes the given Filecoin address into its corresponding public key hash.

        :param address: The Filecoin address to be decoded.
        :type address: str
        :param kwargs: Additional keyword arguments, including "address_type".
        :type kwargs: Any

        :return: The decoded public key hash.
        :rtype: str
        """

        prefix_got: str = address[:len(cls.address_prefix)]
        if cls.address_prefix != prefix_got:
            raise ValueError(f"Invalid prefix (expected: {cls.address_prefix}, got: {prefix_got})")
        address_no_prefix: str = address[len(cls.address_prefix):]

        if not kwargs.get("address_type"):
            address_type: int = cls.address_types[Filecoin.PARAMS.DEFAULT_ADDRESS_TYPE]
        else:
            if kwargs.get("address_type") not in Filecoin.ADDRESS_TYPES.get_address_types():
                raise AddressError(
                    f"Invalid {cls.name()} address type",
                    expected=Filecoin.ADDRESS_TYPES.get_address_types(),
                    got=kwargs.get("address_type")
                )
            address_type: int = cls.address_types[kwargs.get("address_type")]

        address_type_got = ord(address_no_prefix[0]) - ord("0")
        if address_type != address_type_got:
            raise ValueError(f"Invalid address type (expected: {address_type}, got: {address_type_got})")

        address_decode: bytes = get_bytes(decode(address_no_prefix[1:], cls.alphabet))

        if len(address_decode) != 24:
            raise ValueError(f"Invalid length (expected: {24}, got: {len(address_decode)})")

        checksum: bytes = address_decode[-1 * 4:]
        public_key_hash: bytes = address_decode[:-1 * 4]

        checksum_got: bytes = cls.compute_checksum(public_key_hash, address_type)
        if checksum != checksum_got:
            raise ValueError(f"Invalid checksum (expected: {checksum.hex()}, got: {checksum_got.hex()})")

        return bytes_to_string(public_key_hash)



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from ..slip44 import CoinTypes
from ..eccs import SLIP10Ed25519Blake2bECC
from ..consts import (
    Info, Entropies, Mnemonics, Seeds, HDs, Addresses, Networks, Params, XPrivateKeyVersions, XPublicKeyVersions
)
from .icryptocurrency import (
    ICryptocurrency, INetwork
)


class Mainnet(INetwork):

    NAME = "mainnet"
    XPRIVATE_KEY_VERSIONS = XPrivateKeyVersions({
        "P2PKH": 0x0488ade4
    })
    XPUBLIC_KEY_VERSIONS = XPublicKeyVersions({
        "P2PKH": 0x0488b21e
    })


class Nano(ICryptocurrency):

    NAME = "Nano"
    SYMBOL = "XNO"
    INFO = Info({
        "SOURCE_CODE": "https://github.com/nanocurrency/nano-node",
        "WHITEPAPER": "https://nano.org/en/whitepaper",
        "WEBSITES": [
            "http://nano.org/en"
        ]
    })
    ECC = SLIP10Ed25519Blake2bECC
    COIN_TYPE = CoinTypes.Nano
    SUPPORT_BIP38 = False
    NETWORKS = Networks({
        "MAINNET": Mainnet
    })
    DEFAULT_NETWORK = NETWORKS.MAINNET
    ENTROPIES = Entropies({
        "BIP39"
    })
    MNEMONICS = Mnemonics({
        "BIP39"
    })
    SEEDS = Seeds({
        "BIP39"
    })
    HDS = HDs({
        "BIP32", "BIP44"
    })
    DEFAULT_HD = HDS.BIP44
    DEFAULT_PATH = f"m/44'/{COIN_TYPE}'/0'/0/0"
    ADDRESSES = Addresses({
        "NANO": "Nano"
    })
    DEFAULT_ADDRESS = ADDRESSES.NANO
    SEMANTICS = [
        "p2pkh"
    ]
    DEFAULT_SEMANTIC = "p2pkh"
    PARAMS = Params({
        "ADDRESS_PREFIX": "nano_",
        "ALPHABET": "13456789abcdefghijkmnopqrstuwxyz",
        "PAYLOAD_PADDING_DECODED": b"\x00\x00\x00",
        "PAYLOAD_PADDING_ENCODED": "1111"
    })



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Union, Tuple
)

from ..utils import (
    normalize_index, normalize_derivation, index_tuple_to_string
)
from .iderivation import IDerivation


class MoneroDerivation(IDerivation):
    """
    This class implements the Monero standard for hierarchical deterministic wallets.
    Monero defines a specific path structure for deriving keys from a master seed.

    .. note::
        This class inherits from the ``IDerivation`` class, thereby ensuring that all functions are accessible.
    """

    _minor: Union[Tuple[int, bool], Tuple[int, int, bool]]
    _major: Union[Tuple[int, bool], Tuple[int, int, bool]]

    def __init__(
        self, minor: Union[str, int, Tuple[int, int]] = 1, major: Union[str, int, Tuple[int, int]] = 0
    ) -> None:
        """
        Initialize a MoneroDerivation instance with the provided minor and major indexes.

        :param minor: The minor index to set. Can be a string, integer, or tuple of integers. Defaults to 1.
        :type minor: Union[str, int, Tuple[int, int]]
        :param major: The major index to set. Can be a string, integer, or tuple of integers. Defaults to 0.
        :type major: Union[str, int, Tuple[int, int]]

        :return: No return
        :rtype: NoneType
        """

        super(MoneroDerivation, self).__init__()

        self._minor = normalize_index(index=minor, hardened=False)
        self._major = normalize_index(index=major, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._minor)}/"
            f"{index_tuple_to_string(index=self._major)}"
        ))

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the derivation class.

        :return: The name of the derivation class.
        :rtype: str
        """

        return "Monero"

    def from_minor(self, minor: Union[str, int, Tuple[int, int]]) -> "MoneroDerivation":
        """
        Sets the minor index for Monero derivation and updates the derivation path accordingly.

        This method accepts a minor index, normalizes it, and updates the derivation path to reflect the new
        minor index value.

        :param minor: The minor index to set. Can be a string, integer, or tuple of integers.
        :type minor: Union[str, int, Tuple[int, int]]

        :return: The instance of MoneroDerivation with the updated minor index.
        :rtype: MoneroDerivation
        """

        self._minor = normalize_index(index=minor, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._minor)}/"
            f"{index_tuple_to_string(index=self._major)}"
        ))
        return self

    def from_major(self, major: Union[str, int, Tuple[int, int]]) -> "MoneroDerivation":
        """
        Sets the major index for Monero derivation and updates the derivation path accordingly.

        This method accepts a major index, normalizes it, and updates the derivation path to reflect the new
        major index value.

        :param major: The major index to set. Can be a string, integer, or tuple of integers.
        :type major: Union[str, int, Tuple[int, int]]

        :return: The instance of MoneroDerivation with the updated major index.
        :rtype: MoneroDerivation
        """

        self._major = normalize_index(index=major, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._minor)}/"
            f"{index_tuple_to_string(index=self._major)}"
        ))
        return self

    def clean(self) -> "MoneroDerivation":
        """
        Resets the derivation path to default values for Monero derivation.

        This method sets the `minor` index to 1 and the `major` index to 0, and updates the derivation path
        accordingly.

        :return: The instance of MoneroDerivation with the reset derivation path.
        :rtype: MoneroDerivation
        """

        self._minor = normalize_index(index=1, hardened=False)
        self._major = normalize_index(index=0, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._minor)}/"
            f"{index_tuple_to_string(index=self._major)}"
        ))
        return self

    def minor(self) -> int:
        """
        Retrieves the minor index from the internal `_minor` attribute.

        :return: The minor index.
        :rtype: int
        """

        return (
            self._minor[1] if len(self._minor) == 3 else self._minor[0]
        )

    def major(self) -> int:
        """
        Retrieves the major index from the internal `_major` attribute.

        :return: The major index.
        :rtype: int
        """

        return (
            self._major[1] if len(self._major) == 3 else self._major[0]
        )



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from ..slip44 import CoinTypes
from ..eccs import SLIP10Secp256k1ECC
from ..consts import (
    Info, WitnessVersions, Entropies, Mnemonics, Seeds, HDs, Addresses, AddressTypes, Networks, XPrivateKeyVersions, XPublicKeyVersions
)
from .icryptocurrency import (
    ICryptocurrency, INetwork
)


class Mainnet(INetwork):

    NAME = "mainnet"
    LEGACY_PUBLIC_KEY_ADDRESS_PREFIX = 0x00
    LEGACY_SCRIPT_ADDRESS_PREFIX = 0x05
    STD_PUBLIC_KEY_ADDRESS_PREFIX = 0x00
    STD_SCRIPT_ADDRESS_PREFIX = 0x08
    HRP = "bitcoincash"
    WITNESS_VERSIONS = WitnessVersions({
        "P2WPKH": 0x00,
        "P2WSH": 0x00
    })
    XPRIVATE_KEY_VERSIONS = XPrivateKeyVersions({
        "P2PKH": 0x0488ade4,
        "P2SH": 0x0488ade4,
        "P2WPKH": 0x04b2430c,
        "P2WPKH_IN_P2SH": 0x049d7878,
        "P2WSH": 0x02aa7a99,
        "P2WSH_IN_P2SH": 0x0295b005
    })
    XPUBLIC_KEY_VERSIONS = XPublicKeyVersions({
        "P2PKH": 0x0488b21e,
        "P2SH": 0x0488b21e,
        "P2WPKH": 0x04b24746,
        "P2WPKH_IN_P2SH": 0x049d7cb2,
        "P2WSH": 0x02aa7ed3,
        "P2WSH_IN_P2SH": 0x0295b43f
    })
    WIF_PREFIX = 0x80


class Testnet(INetwork):

    NAME = "testnet"
    LEGACY_PUBLIC_KEY_ADDRESS_PREFIX = 0x6f
    LEGACY_SCRIPT_ADDRESS_PREFIX = 0xc4
    STD_PUBLIC_KEY_ADDRESS_PREFIX = 0x00
    STD_SCRIPT_ADDRESS_PREFIX = 0x08
    HRP = "bchtest"
    WITNESS_VERSIONS = WitnessVersions({
        "P2WPKH": 0x00,
        "P2WSH": 0x00
    })
    XPRIVATE_KEY_VERSIONS = XPrivateKeyVersions({
        "P2PKH": 0x04358394,
        "P2SH": 0x04358394,
        "P2WPKH": 0x045f18bc,
        "P2WPKH_IN_P2SH": 0x044a4e28,
        "P2WSH": 0x02575048,
        "P2WSH_IN_P2SH": 0x024285b5
    })
    XPUBLIC_KEY_VERSIONS = XPublicKeyVersions({
        "P2PKH": 0x043587cf,
        "P2SH": 0x043587cf,
        "P2WPKH": 0x045f1cf6,
        "P2WPKH_IN_P2SH": 0x044a5262,
        "P2WSH": 0x02575483,
        "P2WSH_IN_P2SH": 0x024289ef
    })
    WIF_PREFIX = 0xef


class Regtest(Testnet):

    NAME = "regtest"
    HRP = "bchreg"


class BitcoinCash(ICryptocurrency):

    NAME = "Bitcoin-Cash"
    SYMBOL = "BCH"
    INFO = Info({
        "SOURCE_CODE": "https://github.com/bitcoincashorg/bitcoincash.org",
        "WEBSITES": [
            "http://bch.info"
        ]
    })
    ECC = SLIP10Secp256k1ECC
    COIN_TYPE = CoinTypes.BitcoinCash
    SUPPORT_BIP38 = False
    NETWORKS = Networks({
        "MAINNET": Mainnet, "TESTNET": Testnet, "REGTEST": Regtest
    })
    DEFAULT_NETWORK = NETWORKS.MAINNET
    ENTROPIES = Entropies({
        "BIP39"
    })
    MNEMONICS = Mnemonics({
        "BIP39"
    })
    SEEDS = Seeds({
        "BIP39"
    })
    HDS = HDs({
        "BIP32", "BIP44"
    })
    DEFAULT_HD = HDS.BIP44
    DEFAULT_PATH = f"m/44'/{COIN_TYPE}'/0'/0/0"
    ADDRESSES = Addresses((
        "P2PKH", "P2SH", "P2WPKH", {"P2WPKH_IN_P2SH": "P2WPKH-In-P2SH"}, "P2WSH", {"P2WSH_IN_P2SH": "P2WSH-In-P2SH"}
    ))
    DEFAULT_ADDRESS = ADDRESSES.P2PKH
    SEMANTICS = [
        "p2pkh", "p2sh", "p2wpkh", "p2wpkh-in-p2sh", "p2wsh", "p2wsh-in-p2sh"
    ]
    DEFAULT_SEMANTIC = "p2pkh"
    ADDRESS_TYPES = AddressTypes({
        "STD": "std",
        "LEGACY": "legacy"
    })
    DEFAULT_ADDRESS_TYPE = ADDRESS_TYPES.STD



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Union, List, Optional
)

from ..eccs import IEllipticCurveCryptography
from ..derivations import IDerivation


class IHD:

    _ecc: IEllipticCurveCryptography
    _derivation: IDerivation

    def __init__(self, **kwargs) -> None:
        self._ecc = kwargs.get("ecc")

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the hd class.

        :return: The name of the hd class.
        :rtype: str
        """

    def from_seed(self, seed: Union[bytes, str], **kwargs) -> "IHD":
        """
        Initializes the HD instance from the given seed.

        :param seed: The seed to initialize the instance. It can be of type `bytes`, `str`, or `ISeed`.
        :type seed: Union[bytes, str]
        :param kwargs: Additional keyword arguments.

        :return: The initialized HD instance.
        :rtype: HD
        """

    def from_xprivate_key(
        self, xprivate_key: str, encoded: bool = True, strict: bool = False
    ) -> "IHD":
        """
        Initializes the HD instance from the given extended private key (xprivate key).

        :param xprivate_key: The extended private key to initialize the instance.
        :type xprivate_key: str
        :param encoded: Indicates if the xprivate key is encoded. Defaults to True.
        :type encoded: bool
        :param strict: If set to True, enforces strict checking to ensure the xprivate key is a root key. Defaults to False.
        :type strict: bool

        :return: The initialized HD instance.
        :rtype: HD
        """

    def from_xpublic_key(
        self, xpublic_key: str, encoded: bool = True, strict: bool = False
    ) -> "IHD":
        """
        Initializes the HD instance from the given extended public key (xpublic key).

        :param xpublic_key: The extended public key to initialize the instance.
        :type xpublic_key: str
        :param encoded: Indicates if the xpublic key is encoded. Defaults to True.
        :type encoded: bool
        :param strict: If set to True, enforces strict checking to ensure the xpublic key is a root key. Defaults to False.
        :type strict: bool

        :return: The initialized HD instance.
        :rtype: HD
        """

    def from_wif(self, wif: str) -> "IHD":
        """
        Initializes the HD instance from the given Wallet Import Format (WIF) key.

        :param wif: The Wallet Import Format (WIF) key to initialize the instance.
        :type wif: str

        :return: The initialized HD instance.
        :rtype: HD
        """

    def from_private_key(self, private_key: str) -> "IHD":
        """
        Initializes the HD instance from the given private key.

        :param private_key: The private key to initialize the instance, represented as a string.
        :type private_key: str

        :return: The initialized HD instance.
        :rtype: HD
        """

    def from_spend_private_key(self, spend_private_key: str) -> "IHD":
        """
        Initialize the HD instance from a spend private key.

        :param spend_private_key: Spend private key to initialize from.
        :type spend_private_key: str

        :return: Updated HD instance initialized from the spend private key.
        :rtype: HD
        """

    def from_public_key(self, public_key: str) -> "IHD":
        """
        Initializes the HD instance from the given public key.

        :param public_key: The public key to initialize the instance, represented as a string.
        :type public_key: str

        :return: The initialized HD instance.
        :rtype: HD
        """

    def from_watch_only(self, view_private_key, spend_public_key) -> "IHD":
        """
        Initialize the MoneroHD instance from watch-only keys.

        :param view_private_key: View private key or seed to initialize from.
        :param spend_public_key: Spend public key to initialize from.

        :return: Updated HD instance initialized from the watch-only keys.
        :rtype: MoneroHD
        """

    def from_derivation(self, derivation: IDerivation) -> "IHD":
        """
        Initializes the MoneroHD instance using the specified derivation path.

        :param derivation: The derivation path to initialize the instance.
                           It must be an instance of `IDerivation`.
        :type derivation: IDerivation

        :return: The initialized MoneroHD instance.
        :rtype: MoneroHD
        """

    def update_derivation(self, derivation: IDerivation) -> "IHD":
        """
        Updates the derivation path for the HD instance.

        This method first cleans the current derivation path and then sets the new
        derivation path provided. It drives the instance through the new derivation path.

        :param derivation: The new derivation path to set. It must be an instance of `IDerivation`.
        :type derivation: IDerivation

        :return: The updated HD instance.
        :rtype: HD
        """

    def clean_derivation(self) -> "IHD":
        """
        Cleans the derivation path of the HD instance.

        :return: The cleaned HD instance.
        :rtype: HD
        """

    def derivation(self) -> IDerivation:
        """
        Retrieves the derivation method used for key generation.

        :return: An object representing the derivation details.
        :rtype: IDerivation
        """
        return self._derivation

    def seed(self) -> Optional[str]:
        """
        Retrieves the seed value as a string if it exists.

        :return: The seed value as a string, or None if the seed is not set.
        :rtype: Optional[str]
        """

    def semantic(self) -> Optional[str]:
        """
        Retrieves the semantic value as a string if it exists.

        :return: The semantic value as a string, or None if the semantic is not set.
        :rtype: Optional[str]
        """

        return None

    def root_xprivate_key(self, *args, **kwargs) -> Optional[str]:
        """
        Retrieves the root extended private key (xprv) in serialized format.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The root extended private key (xprv) in serialized format, or None if the chain code is not set.
        :rtype: Optional[str]
        """

    def root_xpublic_key(self, *args, **kwargs) -> Optional[str]:
        """
        Retrieves the root extended public key (xpub) in serialized format.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The root extended public key (xpub) in serialized format, or None if the chain code is not set.
        :rtype: Optional[str]
        """

    def master_xprivate_key(self, *args, **kwargs) -> Optional[str]:
        """
        Retrieves the master extended private key in serialized format.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The root extended private key in serialized format, or None if the chain code is not set.
        :rtype: Optional[str]
        """

        return self.root_xprivate_key(args, kwargs)

    def master_xpublic_key(self, *args, **kwargs) -> Optional[str]:
        """
        Retrieves the master extended public key in serialized format.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The master extended public key in serialized format, or None if the chain code is not set.
        :rtype: Optional[str]
        """

        return self.root_xpublic_key(args, kwargs)

    def root_private_key(self, *args, **kwargs) -> Optional[str]:
        """
        Retrieves the root private key as a string.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The root private key as a hexadecimal string, or None if not set.
        :rtype: Optional[str]
        """

    def root_wif(self, *args, **kwargs) -> Optional[str]:
        """
        Retrieves the root private key in WIF format.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The root private key in WIF format as a string, or None if the root private key is not set.
        :rtype: Optional[str]
        """

    def root_chain_code(self) -> Optional[str]:
        """
        Retrieves the root chain code as a string.

        :return: The root chain code as a string, or None if the root chain code is not set.
        :rtype: Optional[str]
        """

    def root_public_key(self, *args, **kwargs) -> Optional[str]:
        """
        Retrieves the root public key as a string.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The root public key as a string, or None if the root public key is not set.
        :rtype: Optional[str]
        """

    def master_private_key(self, *args, **kwargs) -> Optional[str]:
        """
        Retrieves the master private key as a string.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The master private key as a string, or None if the master private key is not set.
        :rtype: Optional[str]
        """

    def master_wif(self, *args, **kwargs) -> Optional[str]:
        """
        Retrieves the master wif as a string.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The master wif as a string, or None if the master private is not set.
        :rtype: Optional[str]
        """

    def master_chain_code(self) -> Optional[str]:
        """
        Retrieves the master chain code as a string.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The master chain code as a string, or None if the master chain code is not set.
        :rtype: Optional[str]
        """

        return self.root_chain_code()

    def master_public_key(self, *args, **kwargs) -> str:
        """
        Retrieves the master public key as a string.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The master public key as a string, or None if the master public key is not set.
        :rtype: Optional[str]
        """

    def xprivate_key(self, *args, **kwargs) -> Optional[str]:
        """
        Retrieves the extended private key (xprivate key) as a serialized string.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The serialized xprivate key as a string, or None if the private key or chain code is not set.
        :rtype: Optional[str]
        """

    def xpublic_key(self, *args, **kwargs) -> Optional[str]:
        """
        Retrieves the extended public key (xpublic key) as a serialized string.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The serialized xpublic key as a string, or None if the chain code is not set.
        :rtype: Optional[str]
        """

    def private_key(self, *args, **kwargs) -> Optional[str]:
        """
        Retrieves the private key as a string.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The private key as a string, or None if the private key is not set.
        :rtype: Optional[str]
        """

    def spend_private_key(self) -> str:
        """
        Retrieves the spend private key used in this MoneroHD instance.

        :return: Spend private key used in this instance, or None if not set.
        :rtype: Optional[str]
        """

    def view_private_key(self) -> str:
        """
        Retrieves the view private key used in this MoneroHD instance.

        :return: View private key used in this instance.
        :rtype: str
        """

    def wif(self, *args, **kwargs) -> Optional[str]:
        """
        Converts the private key to a WIF (Wallet Import Format) string.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The WIF representation of the private key, or None if the private key is not set.
        :rtype: Optional[str]
        """

    def wif_type(self) -> str:
        """
        Retrieves the current WIF (Wallet Import Format) type used by the instance.

        :return: The current WIF type if a WIF is present, otherwise None.
        :rtype: Optional[str]
        """

    def chain_code(self) -> str:
        """
        Retrieves the chain code associated with the current instance.

        :return: The chain code as a string if available, otherwise None.
        :rtype: Optional[str]
        """

    def public_key(self, *args, **kwargs) -> str:
        """
        Retrieves the public key associated with the current instance.

        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: The public key as a string based on the specified type.
        :rtype: str
        """

    def compressed(self) -> str:
        """
        Retrieves the compressed form of the public key associated with the current instance.

        :return: The compressed public key.
        :rtype: str
        """

    def uncompressed(self) -> str:
        """
        Retrieves the uncompressed form of the public key associated with the current instance.

        :return: The uncompressed public key.
        :rtype: str
        """

    def spend_public_key(self) -> str:
        """
        Retrieves the spend public key used in this MoneroHD instance.

        :return: Spend public key used in this instance.
        :rtype: str
        """

    def view_public_key(self) -> str:
        """
        Retrieves the view public key used in this MoneroHD instance.

        :return: View public key used in this instance.
        :rtype: str
        """

    def public_key_type(self) -> str:
        """
        Retrieves the type of public key associated with the current instance.

        :return: The type of public key.
        :rtype: str
        """


    def mode(self) -> str:
        """
        Get the mode of the ElectrumV2HD instance.

        :return: The mode of the instance.
        :rtype: str
        """

    def hash(self) -> str:
        """
        Retrieves the hash of the public key.

        :return: The hash of the public key.
        :rtype: str
        """

    def fingerprint(self) -> str:
        """
        Computes the fingerprint of the HD object by hashing the public key.

        :return: The fingerprint of the HD object.
        :rtype: str
        """

    def parent_fingerprint(self) -> str:
        """
        Retrieves the parent fingerprint of the HD object.

        :return: The parent fingerprint if available, otherwise None.
        :rtype: Optional[str]
        """

    def depth(self) -> int:
        """
        Retrieves the depth of the HD object in the hierarchical tree.

        :return: The depth of the object.
        :rtype: int
        """

    def path(self) -> str:
        """
        Retrieves the derivation path associated with the HD object.

        :return: The derivation path.
        :rtype: str
        """

    def path_key(self) -> str:
        """
        Retrieves the derivation path key associated with the HD object.

        :return: The derivation path.
        :rtype: str
        """

    def index(self) -> int:
        """
        Retrieves the index of the current BIP32HD object.

        :return: The index value.
        :rtype: int
        """

    def indexes(self) -> List[int]:
        """
        Retrieves the list of indexes from the derivation associated with this BIP32HD object.

        :return: The list of indexes.
        :rtype: List[int]
        """

    def strict(self) -> Optional[bool]:
        """
        Retrieves the strict mode flag associated with this BIP32HD object.

        :return: The strict mode flag.
        :rtype: Optional[bool]
        """

    def integrated_address(self, **kwargs) -> str:
        """
        Generates the integrated Monero address associated with the spend and view public keys.
        
        :param kwargs: Additional keyword arguments.

        :return: Integrated Monero address.
        :rtype: str
        """

    def primary_address(self, **kwargs) -> str:
        """
        Generates the primary Monero address associated with the spend and view public keys.
        
        :param kwargs: Additional keyword arguments.

        :return: Primary Monero address.
        :rtype: str
        """

    def sub_address(self, **kwargs) -> str:
        """
        Generates a sub-address associated with the given minor and major indexes or uses the current derivation indexes.

        :param kwargs: Additional keyword arguments.
        
        :return: Generated sub-address.
        :rtype: str
        """

    def address(self, **kwargs) -> str:
        """
        Generates an address based on the specified type and parameters.

        :param kwargs: Additional keyword arguments.

        :return: The generated address.
        :rtype: str
        """


#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from tabulate import tabulate

import click

from ...mnemonics import MNEMONICS


def list_languages():

    for index, mnemonic in enumerate(MNEMONICS.classes()):

        languages: list = []
        for _language in mnemonic.languages:
            language: str = ""
            for index, _ in enumerate(_language.split("-")):
                language += _.title() if index == 0 else f"-{_.title()}"
            languages.append([language])

        click.echo(tabulate(
            languages,
            [
                f"{mnemonic.name()} Languages"
            ],
            stralign="left",
            numalign="left"
        ))
        if index != (len(MNEMONICS.classes()) - 1):
            click.echo("\n")
